"""Common classes and functions that are used across the package."""

from ngio.common._dimensions import Dimensions
from ngio.common._masking_roi import compute_masking_roi
from ngio.common._pyramid import consolidate_pyramid, init_empty_pyramid, on_disk_zoom
from ngio.common._roi import (
    Roi,
    RoiPixels,
)
from ngio.common._zoom import InterpolationOrder, dask_zoom, numpy_zoom

__all__ = [
    "Dimensions",
    "InterpolationOrder",
    "Roi",
    "RoiPixels",
    "compute_masking_roi",
    "consolidate_pyramid",
    "dask_zoom",
    "init_empty_pyramid",
    "numpy_zoom",
    "on_disk_zoom",
]



"""Predefined resources for testing and demonstration purposes."""

from pathlib import Path
from typing import Literal

from ngio.resources.resource_model import LabelsInfo, SampleInfo

resources = Path(__file__).parent.resolve()

_resources = {
    "Cardiomyocyte": SampleInfo(
        img_path=resources
        / "20200812-CardiomyocyteDifferentiation14-Cycle1_B03"
        / "raw.jpg",
        labels=[
            LabelsInfo(
                name="nuclei",
                label_path=resources
                / "20200812-CardiomyocyteDifferentiation14-Cycle1_B03"
                / "nuclei.png",
                create_masking_table=False,
                ensure_unique_labels=True,
            ),
            LabelsInfo(
                name="nuclei_mask",
                label_path=resources
                / "20200812-CardiomyocyteDifferentiation14-Cycle1_B03"
                / "mask.png",
                create_masking_table=True,
                ensure_unique_labels=False,
                dtype="uint8",
            ),
        ],
        xy_pixelsize=0.325,
        z_spacing=1.0,
        time_spacing=1.0,
        name="Cardiomyocyte Differentiation",
        info="20200812-CardiomyocyteDifferentiation14-Cycle1_B03",
    )
}

AVAILABLE_SAMPLES = Literal["Cardiomyocyte"]


def get_sample_info(name: AVAILABLE_SAMPLES) -> SampleInfo:
    """Get a predefined resource by name."""
    image_info = _resources.get(name)
    if image_info is None:
        raise ValueError(
            f"Sample '{name}' not found. Available samples: {_resources.keys()}"
        )
    return image_info


__all__ = ["AVAILABLE_SAMPLES", "LabelsInfo", "SampleInfo"]



"""Utility to read/write OME-Zarr metadata v0.4."""

from ngio.ome_zarr_meta.v05._v05_spec_utils import (
    ngio_to_v05_image_meta,
    ngio_to_v05_label_meta,
    ngio_to_v05_plate_meta,
    ngio_to_v05_well_meta,
    v05_to_ngio_image_meta,
    v05_to_ngio_label_meta,
    v05_to_ngio_plate_meta,
    v05_to_ngio_well_meta,
)

__all__ = [
    "ngio_to_v05_image_meta",
    "ngio_to_v05_label_meta",
    "ngio_to_v05_plate_meta",
    "ngio_to_v05_well_meta",
    "v05_to_ngio_image_meta",
    "v05_to_ngio_label_meta",
    "v05_to_ngio_plate_meta",
    "v05_to_ngio_well_meta",
]



from abc import ABC, abstractmethod
from typing import Literal

from anndata import AnnData
from pandas import DataFrame
from polars import DataFrame as PolarsDataFrame
from polars import LazyFrame
from pydantic import BaseModel, ConfigDict

from ngio.tables.backends._utils import (
    TabularData,
    convert_to_anndata,
    convert_to_pandas,
    convert_to_polars,
)
from ngio.utils import NgioValueError, ZarrGroupHandler


class BackendMeta(BaseModel):
    """Metadata for the backend."""

    backend: str = "anndata_v1"
    index_key: str | None = None
    index_type: Literal["int", "str"] | None = None

    model_config = ConfigDict(extra="allow")


class AbstractTableBackend(ABC):
    """Abstract class for table backends."""

    def set_group_handler(
        self,
        group_handler: ZarrGroupHandler,
        index_key: str | None = None,
        index_type: Literal["int", "str"] | None = None,
    ) -> None:
        """Attach a group handler to the backend.

        Index keys and index types are used to ensure that the
        serialization and deserialization of the table
        is consistent across different backends.

        Making sure that this is consistent is
        a duty of the backend implementations.
        """
        self._group_handler = group_handler
        self._index_key = index_key
        self._index_type = index_type

    @staticmethod
    @abstractmethod
    def backend_name() -> str:
        """Return the name of the backend.

        As a convention we set name as:
            {backend_name}_v{version}

        Where the version is a integer.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def implements_anndata() -> bool:
        """Check if the backend implements the anndata protocol.

        If this is True, the backend should implement the
        `write_from_anndata` method.

        AnnData objects are more complex than DataFrames,
        so if this is true the backend should implement the
        full serialization of the AnnData object.

        If this is False, these methods should raise a
        `NotImplementedError`.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def implements_pandas() -> bool:
        """Check if the backend implements the pandas protocol.

        If this is True, the backend should implement the
        `write_from_dataframe` methods.

        If this is False, these methods should raise a
        `NotImplementedError`.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def implements_polars() -> bool:
        """Check if the backend implements the polars protocol.

        If this is True, the backend should implement the
        `write_from_polars` methods.

        If this is False, these methods should raise a
        `NotImplementedError`.
        """
        raise NotImplementedError

    @property
    def group_handler(self) -> ZarrGroupHandler:
        """Get the group handler."""
        return self._group_handler

    @property
    def index_key(self) -> str | None:
        """Get the index key."""
        return self._index_key

    @property
    def index_type(self) -> Literal["int", "str"] | None:
        """Get the index type."""
        if self._index_type is None:
            return None

        if self._index_type not in ["int", "str"]:
            raise NgioValueError(
                f"Invalid index type {self._index_type}. Must be 'int' or 'str'."
            )
        return self._index_type  # type: ignore[return-value]

    @abstractmethod
    def load(self) -> TabularData:
        """Load the table from the store.

        This is a generic load method.
        Based on the explicit mode or the type of the table,
        it will call the appropriate load method.
        """
        ...

    def load_as_anndata(self) -> AnnData:
        """Load the table as an AnnData object.

        Since the AnnData object is more complex than a DataFrame,
        selecting columns is not implemented, because it is not
        straightforward to do so for an arbitrary AnnData object.
        """
        table = self.load()
        return convert_to_anndata(
            table,
            index_key=self.index_key,
        )

    def load_as_pandas_df(self) -> DataFrame:
        """Load the table as a pandas DataFrame.

        If columns are provided, the table should be filtered
        """
        table = self.load()
        return convert_to_pandas(
            table,
            index_key=self.index_key,
            index_type=self.index_type,
        )

    def load_as_polars_lf(self) -> LazyFrame:
        """Load the table as a polars LazyFrame.

        If columns are provided, the table should be filtered
        """
        table = self.load()
        return convert_to_polars(
            table,
            index_key=self.index_key,
            index_type=self.index_type,
        )

    def write_from_pandas(self, table: DataFrame) -> None:
        """Serialize the table from a pandas DataFrame."""
        raise NotImplementedError(
            f"Backend {self.backend_name()} does not support "
            "serialization of DataFrame objects."
        )

    def write_from_anndata(self, table: AnnData) -> None:
        """Serialize the table from an AnnData object."""
        raise NotImplementedError(
            f"Backend {self.backend_name()} does not support "
            "serialization of AnnData objects."
        )

    def write_from_polars(self, table: PolarsDataFrame | LazyFrame) -> None:
        """Serialize the table from a polars DataFrame or LazyFrame."""
        raise NotImplementedError(
            f"Backend {self.backend_name()} does not support "
            "serialization of Polars objects."
        )

    def write_metadata(self, metadata: dict | None = None) -> None:
        """Write the metadata to the store."""
        if metadata is None:
            metadata = {}

        attrs = self._group_handler.reopen_group().attrs.asdict()
        # This is required by anndata to identify the format
        if "encoding-type" in attrs:
            metadata["encoding-type"] = attrs["encoding-type"]
        if "encoding-version" in attrs:
            metadata["encoding-version"] = attrs["encoding-version"]

        backend_metadata = BackendMeta(
            backend=self.backend_name(),
            index_key=self.index_key,
            index_type=self.index_type,
        ).model_dump(exclude_none=True)
        metadata.update(backend_metadata)
        self._group_handler.write_attrs(metadata)

    def write(
        self,
        table_data: TabularData,
        metadata: dict | None = None,
    ) -> None:
        """Serialize the table to the store, and write the metadata.

        This is a generic write method.
        Based on the explicit mode or the type of the table,
        it will call the appropriate write method.
        """
        if isinstance(table_data, DataFrame):
            self.write_from_pandas(table_data)
        elif isinstance(table_data, AnnData):
            self.write_from_anndata(table_data)
        elif isinstance(table_data, PolarsDataFrame | LazyFrame):
            self.write_from_polars(table_data)
        else:
            raise NgioValueError(
                f"Unsupported table type {type(table_data)}. "
                "Please specify the mode explicitly. "
                "Supported serialization modes are: "
                "'pandas', 'anndata', 'polars'."
            )
        self.write_metadata(metadata)



from math import ceil

import numpy as np


def _center_crop(arr: np.ndarray, target: int, axis: int) -> np.ndarray:
    # Center-crop the array `arr` along dimension `axis` to size `target`.
    # This assumes target < arr.shape[axis].
    n = arr.shape[axis]
    start = (n - target) // 2
    end = start + target
    slc = [slice(None)] * arr.ndim
    slc[axis] = slice(start, end)
    return arr[tuple(slc)]


def _tile_to(
    arr: np.ndarray, target: int, axis: int, label_mode: bool = False
) -> np.ndarray:
    # Tile the array `arr` along dimension `axis` to size `target`.
    # This assumes target > arr.shape[axis].
    n = arr.shape[axis]
    reps = ceil(target / n)

    tiles = []
    flip = False
    max_label = 0
    for _ in range(reps):
        if flip:
            t_arr = np.flip(arr, axis=axis)
        else:
            t_arr = 1 * arr
        if label_mode:
            # Remove duplicate labels
            t_arr = np.where(t_arr > 0, t_arr + max_label, 0)
            max_label = t_arr.max()
        tiles.append(t_arr)
        flip = not flip

    tiled = np.concatenate(tiles, axis=axis)

    slc = [slice(None)] * arr.ndim
    slc[axis] = slice(0, target)
    return tiled[tuple(slc)]


def _fit_to_shape_2d(
    src: np.ndarray, out_shape: tuple[int, int], label_mode: bool = False
) -> np.ndarray:
    """Fit a 2D array to a target shape by center-cropping or tiling as necessary."""
    out_r, out_c = out_shape
    arr = src
    if out_r < arr.shape[0]:
        arr = _center_crop(arr, out_r, axis=0)
    else:
        arr = _tile_to(arr, out_r, axis=0, label_mode=label_mode)

    if out_c < arr.shape[1]:
        arr = _center_crop(arr, out_c, axis=1)
    else:
        arr = _tile_to(arr, out_c, axis=1, label_mode=label_mode)
    return arr


def fit_to_shape(
    arr: np.ndarray, out_shape: tuple[int, ...], ensure_unique_info: bool = False
) -> np.ndarray:
    """Fit a 2D array to a target shape.

    The x,y dimensions of `arr` are fitted to the last two dimensions of
    `out_shape` by center-cropping or tiling as necessary.
    The other dimensions are broadcasted as necessary.

    WARNING: This does not zoom the image, it only crops or tiles it.

    Args:
        arr (np.ndarray): The input 2D array.
        out_shape (tuple[int, ...]): The target shape. Must have at least 2
            and at most 5 dimensions.
        ensure_unique_info (bool, optional): If True, assumes that `arr` is a label
            image and ensures that labels do not overlap when tiling. Defaults to False.

    Returns:
        np.ndarray: The fitted array with shape `out_shape`.
    """
    if len(out_shape) < 2:
        raise ValueError("`out_shape` must contain at least 2 dimensions.")

    if len(out_shape) > 5:
        raise ValueError("`out_shape` must contain at most 5 dimensions.")

    if any(d <= 0 for d in out_shape):
        raise ValueError("`out_shape` must contain positive integers.")

    if arr.ndim != 2:
        raise ValueError("`arr` must be a 2D array.")

    *_, sy, sx = out_shape
    arr = _fit_to_shape_2d(arr, out_shape=(sy, sx), label_mode=ensure_unique_info)
    arr = np.broadcast_to(arr, out_shape)
    return arr



import math
from collections.abc import Sequence

import dask.array as da
import numpy as np

from ngio.common._dimensions import Dimensions
from ngio.common._zoom import (
    InterpolationOrder,
    dask_zoom,
    numpy_zoom,
)
from ngio.io_pipes._ops_axes import AxesOps
from ngio.io_pipes._ops_slices import SlicingOps


class BaseZoomTransform:
    def __init__(
        self,
        input_dimensions: Dimensions,
        target_dimensions: Dimensions,
        order: InterpolationOrder = "nearest",
    ) -> None:
        self._input_dimensions = input_dimensions
        self._target_dimensions = target_dimensions
        self._input_pixel_size = input_dimensions.pixel_size
        self._target_pixel_size = target_dimensions.pixel_size
        self._order: InterpolationOrder = order

    def _normalize_shape(
        self, slice_: slice | int | tuple, scale: float, max_dim: int
    ) -> int:
        if isinstance(slice_, slice):
            _start = slice_.start or 0
            _start_int = math.floor(_start * scale)
            if slice_.stop is not None:
                _stop = slice_.stop * scale
                _stop = min(_stop, max_dim)
            else:
                _stop = max_dim
            _stop_int = math.ceil(_stop)
            target_shape = _stop_int - _start_int

        elif isinstance(slice_, int):
            target_shape = 1
        elif isinstance(slice_, tuple):
            target_shape = len(slice_) * scale
        else:
            raise ValueError(f"Unsupported slice type: {type(slice_)}")
        return math.ceil(target_shape)

    def _compute_zoom_shape(
        self,
        array_shape: Sequence[int],
        axes_ops: AxesOps,
        slicing_ops: SlicingOps,
    ) -> tuple[int, ...]:
        assert len(array_shape) == len(axes_ops.output_axes)

        target_shape = []
        for shape, ax_name in zip(array_shape, axes_ops.output_axes, strict=True):
            ax_type = self._input_dimensions.axes_handler.get_axis(ax_name)
            if ax_type is None:
                # Unknown axis can only be a virtual axis
                # So we set it to 1
                target_shape.append(1)
                continue
            elif ax_type.axis_type == "channel":
                # Do not scale channel axis
                target_shape.append(shape)
                continue
            t_dim = self._target_dimensions.get(ax_name, default=1)
            in_pix = self._input_pixel_size.get(ax_name, default=1.0)
            t_pix = self._target_pixel_size.get(ax_name, default=1.0)
            slice_ = slicing_ops.get(ax_name, normalize=False)
            scale = in_pix / t_pix
            _target_shape = self._normalize_shape(
                slice_=slice_, scale=scale, max_dim=t_dim
            )
            target_shape.append(_target_shape)
        return tuple(target_shape)

    def _compute_inverse_zoom_shape(
        self,
        array_shape: Sequence[int],
        axes_ops: AxesOps,
        slicing_ops: SlicingOps,
    ) -> tuple[int, ...]:
        assert len(array_shape) == len(axes_ops.output_axes)

        target_shape = []
        for shape, ax_name in zip(array_shape, axes_ops.output_axes, strict=True):
            ax_type = self._input_dimensions.axes_handler.get_axis(ax_name)
            if ax_type is not None and ax_type.axis_type == "channel":
                # Do not scale channel axis
                target_shape.append(shape)
                continue
            in_dim = self._input_dimensions.get(ax_name, default=1)
            slice_ = slicing_ops.get(ax_name=ax_name, normalize=True)
            target_shape.append(
                self._normalize_shape(slice_=slice_, scale=1, max_dim=in_dim)
            )

        # Since we are basing the rescaling on the slice, we need to ensure
        # that the input image we got is roughly the right size.
        # This is a safeguard against user errors.
        expected_shape = self._compute_zoom_shape(
            array_shape=target_shape, axes_ops=axes_ops, slicing_ops=slicing_ops
        )
        if any(
            abs(es - s) > 1 for es, s in zip(expected_shape, array_shape, strict=True)
        ):
            raise ValueError(
                f"Input array shape {array_shape} is not compatible with the expected "
                f"shape {expected_shape} based on the zoom transform.\n"
            )
        return tuple(target_shape)

    def _numpy_zoom(
        self, array: np.ndarray, target_shape: tuple[int, ...]
    ) -> np.ndarray:
        if array.shape == target_shape:
            return array
        return numpy_zoom(
            source_array=array, target_shape=target_shape, order=self._order
        )

    def _dask_zoom(
        self,
        array: da.Array,
        array_shape: tuple[int, ...],
        target_shape: tuple[int, ...],
    ) -> da.Array:
        if array_shape == target_shape:
            return array
        return dask_zoom(
            source_array=array, target_shape=target_shape, order=self._order
        )

    def get_as_numpy_transform(
        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> np.ndarray:
        """Apply the scaling transformation to a numpy array."""
        target_shape = self._compute_zoom_shape(
            array_shape=array.shape, axes_ops=axes_ops, slicing_ops=slicing_ops
        )
        return self._numpy_zoom(array=array, target_shape=target_shape)

    def get_as_dask_transform(
        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> da.Array:
        """Apply the scaling transformation to a dask array."""
        array_shape = tuple(int(s) for s in array.shape)
        target_shape = self._compute_zoom_shape(
            array_shape=array_shape, axes_ops=axes_ops, slicing_ops=slicing_ops
        )
        return self._dask_zoom(
            array=array, array_shape=array_shape, target_shape=target_shape
        )

    def set_as_numpy_transform(
        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> np.ndarray:
        """Apply the inverse scaling transformation to a numpy array."""
        target_shape = self._compute_inverse_zoom_shape(
            array_shape=array.shape, axes_ops=axes_ops, slicing_ops=slicing_ops
        )
        return self._numpy_zoom(array=array, target_shape=target_shape)

    def set_as_dask_transform(
        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> da.Array:
        """Apply the inverse scaling transformation to a dask array."""
        array_shape = tuple(int(s) for s in array.shape)
        target_shape = self._compute_inverse_zoom_shape(
            array_shape=array_shape, axes_ops=axes_ops, slicing_ops=slicing_ops
        )
        return self._dask_zoom(
            array=array, array_shape=array_shape, target_shape=target_shape
        )



from pathlib import Path

import pytest
from pandas import DataFrame

from ngio.tables._tables_container import (
    FeatureTable,
    TablesContainer,
    open_tables_container,
)
from ngio.utils import NgioValueError


def test_table_container(tmp_path: Path):
    table_group = open_tables_container(tmp_path / "test.zarr")
    assert isinstance(table_group, TablesContainer)
    assert table_group.list() == []

    # Create a feature table
    table = FeatureTable(
        table_data=DataFrame({"label": [1, 2, 3], "a": [1.0, 1.3, 0.0]})
    )
    table_group.add(name="feat_table", table=table)
    assert table_group.list() == ["feat_table"]

    with pytest.raises(NgioValueError):
        table_group.add(name="feat_table", table=table)

    table = table_group.get("feat_table")
    assert isinstance(table, FeatureTable)

    expected = DataFrame({"label": [1, 2, 3], "a": [1.0, 1.3, 0.0]})
    expected = expected.set_index("label")
    assert table.dataframe.equals(expected)



"""Mappers for iterators.

Mappers are classes that can be passed to the `map` method of iterators to
transform the items yielded by the iterator.

"""

from collections.abc import Callable, Iterable
from typing import Generic, Protocol, TypeVar

from ngio.io_pipes._io_pipes_types import DataGetterProtocol, DataSetterProtocol
from ngio.utils import NgioValueError

T = TypeVar("T")


class MapperProtocol(Protocol[T]):
    """Protocol for mappers."""

    def __call__(
        self,
        func: Callable[[T], T],
        getters: Iterable[DataGetterProtocol[T]],
        setters: Iterable[DataSetterProtocol[T] | None],
    ) -> None:
        """Map an item to another item."""
        ...


class BasicMapper(Generic[T]):
    """A basic mapper that simply applies a function to the data."""

    def __call__(
        self,
        func: Callable[[T], T],
        getters: Iterable[DataGetterProtocol[T]],
        setters: Iterable[DataSetterProtocol[T] | None],
    ) -> None:
        """Map an item to another item."""
        for getter, setter in zip(getters, setters, strict=True):
            data = getter()
            data = func(data)
            if setter is None:
                raise NgioValueError(
                    "Error in BasicMapper: setter is None, "
                    "this iterator is read-only, mapping is not possible."
                )
            setter(data)



import os
import shutil
from pathlib import Path

import pytest

from ngio.utils import download_ome_zarr_dataset

zenodo_download_dir = Path(__file__).parent.parent / "data"
os.makedirs(zenodo_download_dir, exist_ok=True)
cardiomyocyte_tiny_source_path = download_ome_zarr_dataset(
    "CardiomyocyteTiny", download_dir=zenodo_download_dir
)

cardiomyocyte_small_mip_source_path = download_ome_zarr_dataset(
    "CardiomyocyteSmallMip", download_dir=zenodo_download_dir
)


@pytest.fixture
def cardiomyocyte_tiny_path(tmp_path: Path) -> Path:
    dest_path = tmp_path / cardiomyocyte_tiny_source_path.stem
    shutil.copytree(cardiomyocyte_tiny_source_path, dest_path, dirs_exist_ok=True)
    return dest_path


@pytest.fixture
def cardiomyocyte_small_mip_path(tmp_path: Path) -> Path:
    dest_path = tmp_path / cardiomyocyte_small_mip_source_path.stem
    shutil.copytree(cardiomyocyte_small_mip_source_path, dest_path, dirs_exist_ok=True)
    return dest_path


@pytest.fixture
def images_v04(tmp_path: Path) -> dict[str, Path]:
    source = Path("tests/data/v04/images/")
    dest = tmp_path / "v04" / "images"
    dest.mkdir(parents=True, exist_ok=True)
    shutil.copytree(source, dest, dirs_exist_ok=True)
    return {file.name: file for file in dest.glob("*.zarr")}



import fsspec.implementations.http
from aiohttp import ClientResponseError

from ngio.utils import NgioValueError


def fractal_fsspec_store(
    url: str, fractal_token: str | None = None, client_kwargs: dict | None = None
) -> fsspec.mapping.FSMap:
    """Simple function to get an http fsspec store from a url."""
    client_kwargs = {} if client_kwargs is None else client_kwargs
    if fractal_token is not None:
        client_kwargs["headers"] = {"Authorization": f"Bearer {fractal_token}"}
    fs = fsspec.implementations.http.HTTPFileSystem(client_kwargs=client_kwargs)

    store = fs.get_mapper(url)

    possible_keys = [".zgroup", ".zarray"]
    for key in possible_keys:
        try:
            value = store.get(key)
            if value is not None:
                break
        except ClientResponseError as e:
            if e.status == 401 and fractal_token is None:
                raise NgioValueError(
                    "No auto token is provided. You need a valid "
                    f"'fractal_token' to access: {url}."
                ) from e
            elif e.status == 401 and fractal_token is not None:
                raise NgioValueError(
                    f"The 'fractal_token' provided is invalid for: {url}."
                ) from e
            else:
                raise e
    else:
        raise NgioValueError(
            f"Store {url} can not be read. Possible problems are: \n"
            "- The url does not exist. \n"
            "- The url is not a valid .zarr. \n"
        )
    return store



from collections.abc import Sequence

import dask.array as da
import numpy as np

from ngio.common import Roi
from ngio.experimental.iterators._abstract_iterator import AbstractIteratorBuilder
from ngio.images import Image, Label
from ngio.images._image import (
    ChannelSlicingInputType,
    add_channel_selection_to_slicing_dict,
)
from ngio.images._masked_image import MaskedImage
from ngio.io_pipes import (
    DaskGetterMasked,
    DaskRoiGetter,
    DaskRoiSetter,
    DaskSetterMasked,
    NumpyGetterMasked,
    NumpyRoiGetter,
    NumpyRoiSetter,
    NumpySetterMasked,
    TransformProtocol,
)
from ngio.io_pipes._io_pipes_types import DataGetterProtocol, DataSetterProtocol


class SegmentationIterator(AbstractIteratorBuilder[np.ndarray, da.Array]):
    """Base class for iterators over ROIs."""

    def __init__(
        self,
        input_image: Image,
        output_label: Label,
        channel_selection: ChannelSlicingInputType = None,
        axes_order: Sequence[str] | None = None,
        input_transforms: Sequence[TransformProtocol] | None = None,
        output_transforms: Sequence[TransformProtocol] | None = None,
    ) -> None:
        """Initialize the iterator with a ROI table and input/output images.

        Args:
            input_image (Image): The input image to be used as input for the
                segmentation.
            output_label (Label): The label image where the ROIs will be written.
            channel_selection (ChannelSlicingInputType): Optional
                selection of channels to use for the segmentation.
            axes_order (Sequence[str] | None): Optional axes order for the
                segmentation.
            input_transforms (Sequence[TransformProtocol] | None): Optional
                transforms to apply to the input image.
            output_transforms (Sequence[TransformProtocol] | None): Optional
                transforms to apply to the output label.
        """
        self._input = input_image
        self._output = output_label
        self._ref_image = input_image
        self._rois = input_image.build_image_roi_table(name=None).rois()

        # Set iteration parameters
        self._input_slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self._input, channel_selection=channel_selection, slicing_dict={}
        )
        self._channel_selection = channel_selection
        self._axes_order = axes_order
        self._input_transforms = input_transforms
        self._output_transforms = output_transforms

        self._input.require_dimensions_match(self._output, allow_singleton=False)

    def get_init_kwargs(self) -> dict:
        """Return the initialization arguments for the iterator."""
        return {
            "input_image": self._input,
            "output_label": self._output,
            "channel_selection": self._channel_selection,
            "axes_order": self._axes_order,
            "input_transforms": self._input_transforms,
            "output_transforms": self._output_transforms,
        }

    def build_numpy_getter(self, roi: Roi) -> DataGetterProtocol[np.ndarray]:
        return NumpyRoiGetter(
            zarr_array=self._input.zarr_array,
            dimensions=self._input.dimensions,
            roi=roi,
            axes_order=self._axes_order,
            transforms=self._input_transforms,
            slicing_dict=self._input_slicing_kwargs,
        )

    def build_numpy_setter(self, roi: Roi) -> DataSetterProtocol[np.ndarray]:
        return NumpyRoiSetter(
            zarr_array=self._output.zarr_array,
            dimensions=self._output.dimensions,
            roi=roi,
            axes_order=self._axes_order,
            transforms=self._output_transforms,
            remove_channel_selection=True,
        )

    def build_dask_getter(self, roi: Roi) -> DataGetterProtocol[da.Array]:
        return DaskRoiGetter(
            zarr_array=self._input.zarr_array,
            dimensions=self._input.dimensions,
            roi=roi,
            axes_order=self._axes_order,
            transforms=self._input_transforms,
            slicing_dict=self._input_slicing_kwargs,
        )

    def build_dask_setter(self, roi: Roi) -> DataSetterProtocol[da.Array]:
        return DaskRoiSetter(
            zarr_array=self._output.zarr_array,
            dimensions=self._output.dimensions,
            roi=roi,
            axes_order=self._axes_order,
            transforms=self._output_transforms,
            remove_channel_selection=True,
        )

    def post_consolidate(self):
        self._output.consolidate()


class MaskedSegmentationIterator(SegmentationIterator):
    """Base class for iterators over ROIs."""

    def __init__(
        self,
        input_image: MaskedImage,
        output_label: Label,
        channel_selection: ChannelSlicingInputType = None,
        axes_order: Sequence[str] | None = None,
        input_transforms: Sequence[TransformProtocol] | None = None,
        output_transforms: Sequence[TransformProtocol] | None = None,
    ) -> None:
        """Initialize the iterator with a ROI table and input/output images.

        Args:
            input_image (MaskedImage): The input image to be used as input for the
                segmentation.
            output_label (Label): The label image where the ROIs will be written.
            channel_selection (ChannelSlicingInputType): Optional
                selection of channels to use for the segmentation.
            axes_order (Sequence[str] | None): Optional axes order for the
                segmentation.
            input_transforms (Sequence[TransformProtocol] | None): Optional
                transforms to apply to the input image.
            output_transforms (Sequence[TransformProtocol] | None): Optional
                transforms to apply to the output label.
        """
        self._input = input_image
        self._output = output_label

        self._ref_image = input_image
        self._set_rois(input_image._masking_roi_table.rois())

        # Set iteration parameters
        self._input_slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self._input, channel_selection=channel_selection, slicing_dict={}
        )
        self._channel_selection = channel_selection
        self._axes_order = axes_order
        self._input_transforms = input_transforms
        self._output_transforms = output_transforms

        # Check compatibility between input and output images
        # if not self._input.dimensions.is_compatible_with(self._output.dimensions):
        #    raise NgioValidationError(
        #        "Input image and output label have incompatible dimensions. "
        #        f"Input: {self._input.dimensions}, Output: {self._output.dimensions}."
        #    )

    def get_init_kwargs(self) -> dict:
        """Return the initialization arguments for the iterator."""
        return {
            "input_image": self._input,
            "output_label": self._output,
            "channel_selection": self._channel_selection,
            "axes_order": self._axes_order,
            "input_transforms": self._input_transforms,
            "output_transforms": self._output_transforms,
        }

    def build_numpy_getter(self, roi: Roi):
        return NumpyGetterMasked(
            zarr_array=self._input.zarr_array,
            dimensions=self._input.dimensions,
            roi=roi,
            label_zarr_array=self._input._label.zarr_array,
            label_dimensions=self._input._label.dimensions,
            axes_order=self._axes_order,
            transforms=self._input_transforms,
            slicing_dict=self._input_slicing_kwargs,
        )

    def build_numpy_setter(self, roi: Roi):
        return NumpySetterMasked(
            roi=roi,
            zarr_array=self._output.zarr_array,
            dimensions=self._output.dimensions,
            label_zarr_array=self._input._label.zarr_array,
            label_dimensions=self._input._label.dimensions,
            axes_order=self._axes_order,
            transforms=self._output_transforms,
            remove_channel_selection=True,
        )

    def build_dask_getter(self, roi: Roi):
        return DaskGetterMasked(
            roi=roi,
            zarr_array=self._input.zarr_array,
            dimensions=self._input.dimensions,
            label_zarr_array=self._input._label.zarr_array,
            label_dimensions=self._input._label.dimensions,
            axes_order=self._axes_order,
            transforms=self._input_transforms,
            slicing_dict=self._input_slicing_kwargs,
        )

    def build_dask_setter(self, roi: Roi):
        return DaskSetterMasked(
            roi=roi,
            zarr_array=self._output.zarr_array,
            dimensions=self._output.dimensions,
            label_zarr_array=self._input._label.zarr_array,
            label_dimensions=self._input._label.dimensions,
            axes_order=self._axes_order,
            transforms=self._output_transforms,
            remove_channel_selection=True,
        )

    def post_consolidate(self):
        self._output.consolidate()



import io
from collections.abc import Callable
from typing import Any

from pandas import DataFrame
from polars import DataFrame as PolarsDataFrame
from polars import LazyFrame
from zarr.storage import FsspecStore, LocalStore

from ngio.tables.backends._abstract_backend import AbstractTableBackend
from ngio.tables.backends._utils import normalize_pandas_df, normalize_polars_lf
from ngio.utils import NgioFileNotFoundError, NgioValueError


class NonZarrBaseBackend(AbstractTableBackend):
    """A class to load and write small tables in CSV format."""

    def __init__(
        self,
        df_reader: Callable[[Any], DataFrame],
        lf_reader: Callable[[Any], LazyFrame],
        df_writer: Callable[[str, DataFrame], None],
        lf_writer: Callable[[str, PolarsDataFrame], None],
        table_name: str,
    ):
        self.df_reader = df_reader
        self.lf_reader = lf_reader
        self.df_writer = df_writer
        self.lf_writer = lf_writer
        self.table_name = table_name

    @staticmethod
    def implements_anndata() -> bool:
        """Whether the handler implements the anndata protocol."""
        return False

    @staticmethod
    def implements_pandas() -> bool:
        """Whether the handler implements the dataframe protocol."""
        return True

    @staticmethod
    def implements_polars() -> bool:
        """Whether the handler implements the polars protocol."""
        return True

    @staticmethod
    def backend_name() -> str:
        """Return the name of the backend."""
        raise NotImplementedError(
            "The backend_name method must be implemented in the subclass."
        )

    def _load_from_directory_store(self, reader):
        """Load the table from a directory store."""
        url = self._group_handler.full_url
        if url is None:
            ext = self.table_name.split(".")[-1]
            raise NgioValueError(
                f"Ngio does not support reading a {ext} table from a "
                f"store of type {type(self._group_handler)}. "
                "Please make sure to use a compatible "
                "store like a zarr.DirectoryStore."
            )
        table_path = f"{url}/{self.table_name}"
        dataframe = reader(table_path)
        return dataframe

    def _load_from_fs_store_df(self, reader):
        """Load the table from an FS store."""
        path = self._group_handler.group.path
        table_path = f"{path}/{self.table_name}"
        bytes_table = self._group_handler.store.get(table_path)
        if bytes_table is None:
            raise NgioFileNotFoundError(f"No table found at {table_path}. ")
        dataframe = reader(io.BytesIO(bytes_table))
        return dataframe

    def _load_from_fs_store_lf(self, reader):
        """Load the table from an FS store."""
        full_url = self._group_handler.full_url
        parquet_path = f"{full_url}/{self.table_name}"
        store_fs = self._group_handler.store.fs  # type: ignore (in this context, store_fs is a fs.FSStore)
        with store_fs.open(parquet_path, "rb") as f:
            dataframe = reader(f)
        return dataframe

    def load_as_pandas_df(self) -> DataFrame:
        """Load the table as a pandas DataFrame."""
        store = self._group_handler.store
        if isinstance(store, LocalStore):
            dataframe = self._load_from_directory_store(reader=self.df_reader)
        elif isinstance(store, FsspecStore):
            dataframe = self._load_from_fs_store_df(reader=self.df_reader)
        else:
            ext = self.table_name.split(".")[-1]
            raise NgioValueError(
                f"Ngio does not support reading a {ext} table from a "
                f"store of type {type(store)}. "
                "Please make sure to use a compatible "
                "store like a zarr.DirectoryStore or "
                "zarr.FSStore."
            )

        dataframe = normalize_pandas_df(
            dataframe,
            index_key=self.index_key,
            index_type=self.index_type,
            reset_index=False,
        )
        return dataframe

    def load(self) -> DataFrame:
        """Load the table as a pandas DataFrame."""
        return self.load_as_pandas_df()

    def load_as_polars_lf(self) -> LazyFrame:
        """Load the table as a polars LazyFrame."""
        store = self._group_handler.store
        if isinstance(store, LocalStore):
            lazy_frame = self._load_from_directory_store(reader=self.lf_reader)
        elif isinstance(store, FsspecStore):
            lazy_frame = self._load_from_fs_store_lf(reader=self.lf_reader)
        else:
            ext = self.table_name.split(".")[-1]
            raise NgioValueError(
                f"Ngio does not support reading a {ext} from a "
                f"store of type {type(store)}. "
                "Please make sure to use a compatible "
                "store like a zarr.DirectoryStore or "
                "zarr.FSStore."
            )
        if not isinstance(lazy_frame, LazyFrame):
            raise NgioValueError(
                "Table is not a lazy frame. Please report this issue as an ngio bug."
                f" {type(lazy_frame)}"
            )

        lazy_frame = normalize_polars_lf(
            lazy_frame,
            index_key=self.index_key,
            index_type=self.index_type,
        )
        return lazy_frame

    def _get_store_url(self) -> str:
        """Get the store URL."""
        store = self._group_handler.store
        if isinstance(store, LocalStore):
            full_url = self._group_handler.full_url
        else:
            ext = self.table_name.split(".")[-1]
            raise NgioValueError(
                f"Ngio does not support writing a {ext} file to a "
                f"store of type {type(store)}. "
                "Please make sure to use a compatible "
                "store like a zarr.DirectoryStore or "
                "zarr.FSStore."
            )
        if full_url is None:
            ext = self.table_name.split(".")[-1]
            raise NgioValueError(
                f"Ngio does not support writing a {ext} file to a "
                f"store of type {type(store)}. "
                "Please make sure to use a compatible "
                "store like a zarr.DirectoryStore or "
                "zarr.FSStore."
            )
        return full_url

    def write_from_pandas(self, table: DataFrame) -> None:
        """Write the table from a pandas DataFrame."""
        table = normalize_pandas_df(
            table,
            index_key=self.index_key,
            index_type=self.index_type,
            reset_index=True,
        )
        full_url = self._get_store_url()
        table_path = f"{full_url}/{self.table_name}"
        self.df_writer(table_path, table)

    def write_from_polars(self, table: PolarsDataFrame | LazyFrame) -> None:
        """Write the table from a polars DataFrame or LazyFrame."""
        table = normalize_polars_lf(
            table,
            index_key=self.index_key,
            index_type=self.index_type,
        )

        if isinstance(table, LazyFrame):
            table = table.collect()

        full_url = self._get_store_url()
        table_path = f"{full_url}/{self.table_name}"
        self.lf_writer(table_path, table)



"""Fractal internal module for dataset metadata handling."""

import math
from functools import total_ordering
from typing import overload

import numpy as np
from pydantic import BaseModel

from ngio.ome_zarr_meta.ngio_specs import (
    DefaultSpaceUnit,
    DefaultTimeUnit,
    SpaceUnits,
    TimeUnits,
)

################################################################################################
#
# PixelSize model
# The PixelSize model is used to store the pixel size in 3D space.
# The model does not store scaling factors and units for other axes.
#
#################################################################################################


@total_ordering
class PixelSize(BaseModel):
    """PixelSize class to store the pixel size in 3D space."""

    x: float
    y: float
    z: float
    t: float = 1
    space_unit: SpaceUnits | str | None = DefaultSpaceUnit
    time_unit: TimeUnits | str | None = DefaultTimeUnit

    def __repr__(self) -> str:
        """Return a string representation of the pixel size."""
        return f"PixelSize(x={self.x}, y={self.y}, z={self.z}, t={self.t})"

    def __eq__(self, other) -> bool:
        """Check if two pixel sizes are equal."""
        if not isinstance(other, PixelSize):
            raise TypeError("Can only compare PixelSize with PixelSize.")

        if (
            self.time_unit is not None
            and other.time_unit is None
            and self.time_unit != other.time_unit
        ):
            return False

        if self.space_unit != other.space_unit:
            return False
        return math.isclose(self.distance(other), 0)

    def __lt__(self, other: "PixelSize") -> bool:
        """Check if one pixel size is less than the other."""
        if not isinstance(other, PixelSize):
            raise TypeError("Can only compare PixelSize with PixelSize.")
        ref = PixelSize(
            x=0,
            y=0,
            z=0,
            t=0,
            space_unit=self.space_unit,
            time_unit=self.time_unit,  # type: ignore
        )
        return self.distance(ref) < other.distance(ref)

    def as_dict(self) -> dict[str, float]:
        """Return the pixel size as a dictionary."""
        return {"t": self.t, "z": self.z, "y": self.y, "x": self.x}

    @overload
    def get(self, axis: str, default: float) -> float: ...

    @overload
    def get(self, axis: str, default: None = None) -> float | None: ...

    def get(self, axis: str, default: float | None = None) -> float | None:
        """Get the pixel size for a given axis (in canonical name)."""
        px_size = self.as_dict().get(axis, default)
        if px_size is None:
            raise ValueError(
                f"Invalid axis name: {axis}, must be one of 'x', 'y', 'z', 't'."
            )
        return px_size

    @property
    def tzyx(self) -> tuple[float, float, float, float]:
        """Return the voxel size in t, z, y, x order."""
        return self.t, self.z, self.y, self.x

    @property
    def zyx(self) -> tuple[float, float, float]:
        """Return the voxel size in z, y, x order."""
        return self.z, self.y, self.x

    @property
    def yx(self) -> tuple[float, float]:
        """Return the xy plane pixel size in y, x order."""
        return self.y, self.x

    @property
    def voxel_volume(self) -> float:
        """Return the volume of a voxel."""
        return self.y * self.x * self.z

    @property
    def xy_plane_area(self) -> float:
        """Return the area of the xy plane."""
        return self.y * self.x

    @property
    def time_spacing(self) -> float | None:
        """Return the time spacing."""
        return self.t

    def distance(self, other: "PixelSize") -> float:
        """Return the distance between two pixel sizes."""
        return float(np.linalg.norm(np.array(self.tzyx) - np.array(other.tzyx)))



import sys
from pathlib import Path

import dask
import dask.delayed
import fsspec.implementations.http
import numpy as np
import pytest
import zarr
from zarr.storage import LocalStore

from ngio.utils import (
    NgioFileExistsError,
    NgioFileNotFoundError,
    NgioValueError,
    ZarrGroupHandler,
    open_group_wrapper,
)


@pytest.mark.parametrize("cache", [True, False])
def test_group_handler_creation(tmp_path: Path, cache: bool):
    store = tmp_path / "test_group_handler_creation.zarr"
    handler = ZarrGroupHandler(store=store, cache=cache, mode="a")

    _store = handler.group.store
    assert isinstance(_store, LocalStore)
    assert Path(_store.root.as_posix()) == store
    assert handler.use_cache == cache

    attrs = handler.load_attrs()
    assert attrs == {}
    attrs = {"a": 1, "b": 2, "c": 3}
    handler.write_attrs(attrs)
    assert handler.load_attrs() == attrs
    if cache:
        assert handler.get_from_cache("attrs") == attrs
    handler.clean_cache()
    assert handler.get_from_cache("attrs") is None

    handler.write_attrs({"a": 2}, overwrite=False)
    assert handler.load_attrs()["a"] == 2
    assert handler.load_attrs()["b"] == 2

    handler.write_attrs({"a": 3}, overwrite=True)
    assert handler.load_attrs()["a"] == 3
    assert "b" not in handler.load_attrs()

    new_group = handler.create_group("new_group")

    assert isinstance(new_group, zarr.Group)
    assert isinstance(handler.get_group("new_group"), zarr.Group)

    with pytest.raises(NgioFileExistsError):
        handler.create_group("new_group", overwrite=False)


def test_group_handler_from_group(tmp_path: Path):
    store = tmp_path / "test_group_handler_from_group.zarr"
    group = zarr.group(store=store, overwrite=True)

    handler = ZarrGroupHandler(store=group, cache=True, mode="a")
    assert handler.group == group


def test_group_handler_read(tmp_path: Path):
    store = tmp_path / "test_group_handler_read.zarr"

    group = zarr.create_group(store=store, overwrite=True)
    input_attrs = {"a": 1, "b": 2, "c": 3}
    group.attrs.update(input_attrs)

    group.create_group("group1")
    group.create_array("array1", shape=(10, 10), dtype="int32")

    handler = ZarrGroupHandler(store=store, cache=True, mode="r")

    assert handler.load_attrs() == input_attrs
    assert isinstance(handler.get_array("array1"), zarr.Array)
    assert isinstance(handler.get_group("group1"), zarr.Group)
    assert handler.mode == "r"

    with pytest.raises(NgioFileNotFoundError):
        handler.get_array("array2")

    with pytest.raises(NgioFileNotFoundError):
        handler.get_group("group2")

    with pytest.raises(NgioValueError):
        handler.get_array("group1")

    with pytest.raises(NgioValueError):
        handler.get_group("array1")

    with pytest.raises(NgioValueError):
        handler.write_attrs({"a": 1, "b": 2, "c": 3})


def test_open_fail(tmp_path: Path):
    store = tmp_path / "test_open_fail.zarr"
    group = zarr.create_group(store=store, overwrite=True)

    read_only_group = open_group_wrapper(store=group, mode="r")
    assert read_only_group.read_only

    with pytest.raises(NgioFileExistsError):
        open_group_wrapper(store=store, mode="w-")

    with pytest.raises(NgioFileNotFoundError):
        open_group_wrapper(store=store / "non_existent.zarr", mode="r")

    with pytest.raises(NgioValueError):
        open_group_wrapper(store=read_only_group, mode="w")


def test_multiprocessing_safety(tmp_path: Path):
    zarr_store = tmp_path / "test_multiprocessing_safety.zarr"

    @dask.delayed  # type: ignore
    def add_item(i):
        handler = ZarrGroupHandler(
            zarr_store, cache=False, mode="a", parallel_safe=True
        )
        assert handler.lock is not None

        with handler.lock:
            attrs = handler.load_attrs()
            attrs["test_list"].append(i)
            handler.write_attrs(attrs, overwrite=False)

        return i

    handler = ZarrGroupHandler(zarr_store, cache=False, mode="w", parallel_safe=True)
    attrs = handler.load_attrs()
    attrs = {"test_list": []}
    handler.write_attrs(attrs, overwrite=True)

    results = []
    num_items = 1000
    for i in range(num_items):
        results.append(add_item(i))

    dask.compute(*results)  # type: ignore

    _, counts = np.unique(handler.load_attrs()["test_list"], return_counts=True)
    assert len(counts) == num_items
    assert np.all(counts == 1)

    assert handler._lock_path is not None

    if sys.platform.startswith("win"):
        # The file lock path is not removed on Windows
        # for some reason path.exists() returns False
        # even though the file should exist (or at least it does on Mac/Linux)
        return None

    assert Path(handler._lock_path).exists()
    lock_path = Path(handler._lock_path)
    handler.remove_lock()
    assert not lock_path.exists()
    handler.remove_lock()

    handler = ZarrGroupHandler(zarr_store, cache=False, mode="w", parallel_safe=True)
    assert handler.lock is not None
    with pytest.raises(NgioValueError):
        with handler.lock:
            handler.remove_lock()


def test_remote_storage():
    url = (
        "https://raw.githubusercontent.com/"
        "fractal-analytics-platform/fractal-ome-zarr-examples/"
        "refs/heads/main/v04/"
        "20200812-CardiomyocyteDifferentiation14-Cycle1_B_03_mip.zarr/"
    )

    fs = fsspec.implementations.http.HTTPFileSystem(client_kwargs={})
    store = fs.get_mapper(url)
    handler = ZarrGroupHandler(store=store, cache=True, mode="r")
    assert handler.load_attrs()
    assert isinstance(handler.get_array("0"), zarr.Array)
    assert isinstance(handler.get_group("labels"), zarr.Group)

    # Check if the fsspec store based group is handled correctly
    open_group_wrapper(store=handler.group, mode="r")

    with pytest.raises(NgioValueError):
        ZarrGroupHandler(store=store, parallel_safe=True)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import zarr
from anndata import AnnData
from anndata._io.specs import read_elem
from anndata._io.utils import _read_legacy_raw
from anndata._io.zarr import read_dataframe
from anndata.compat import _clean_uns
from anndata.experimental import read_dispatched
from zarr.storage import LocalStore

from ngio.utils import (
    NgioValueError,
    StoreOrGroup,
    open_group_wrapper,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence


def custom_anndata_read_zarr(
    store: StoreOrGroup, elem_to_read: Sequence[str] | None = None
) -> AnnData:
    """Read from a hierarchical Zarr array store.

    # Implementation originally from https://github.com/scverse/anndata/blob/main/src/anndata/_io/zarr.py
    # Original implementation would not work with remote storages so we had to copy it
    # here and slightly modified it to work with remote storages.

    Args:
        store (StoreOrGroup): A store or group to read the AnnData from.
        elem_to_read (Sequence[str] | None): The elements to read from the store.
    """
    group = open_group_wrapper(store=store, mode="r")

    if not isinstance(group.store, LocalStore):
        elem_to_read = ["X", "obs", "var"]

    if elem_to_read is None:
        elem_to_read = [
            "X",
            "obs",
            "var",
            "uns",
            "obsm",
            "varm",
            "obsp",
            "varp",
            "layers",
        ]

    # Read with handling for backwards compat
    def callback(func: Callable, elem_name: str, elem: Any, iospec: Any) -> Any:
        if iospec.encoding_type == "anndata" or elem_name.endswith("/"):
            ad_kwargs = {}
            # Some of these elem fail on https
            # So we only include the ones that are strictly necessary
            # for fractal tables
            # This fails on some https
            # base_elem += list(elem.keys())
            for k in elem_to_read:
                v = elem.get(k)
                if v is not None and not k.startswith("raw."):
                    ad_kwargs[k] = read_dispatched(v, callback)  # type: ignore
            return AnnData(**ad_kwargs)

        elif elem_name.startswith("/raw."):
            return None
        elif elem_name in {"/obs", "/var"}:
            return read_dataframe(elem)
        elif elem_name == "/raw":
            # Backwards compat
            return _read_legacy_raw(group, func(elem), read_dataframe, func)
        return func(elem)

    adata = read_dispatched(group, callback=callback)  # type: ignore

    # Backwards compat (should figure out which version)
    if "raw.X" in group:
        raw = AnnData(**_read_legacy_raw(group, adata.raw, read_dataframe, read_elem))  # type: ignore
        raw.obs_names = adata.obs_names  # type: ignore
        adata.raw = raw  # type: ignore

    # Backwards compat for <0.7
    if isinstance(group["obs"], zarr.Array):
        _clean_uns(adata)

    if isinstance(adata, dict):
        adata = AnnData(**adata)
    if not isinstance(adata, AnnData):
        raise NgioValueError(f"Expected an AnnData object, but got {type(adata)}")
    return adata



"""Fractal internal module for axes handling."""

from collections.abc import Sequence
from enum import Enum
from typing import Literal, TypeAlias, TypeVar

from pydantic import BaseModel, ConfigDict, Field

from ngio.utils import NgioValidationError, NgioValueError

T = TypeVar("T")
SlicingType: TypeAlias = slice | tuple[int, ...] | int

################################################################################################
#
# Axis Types and Units
# We define a small set of axis types and units that can be used in the metadata.
# This axis types are more restrictive than the OME standard.
# We do that to simplify the data processing.
#
#################################################################################################


class AxisType(str, Enum):
    """Allowed axis types."""

    channel = "channel"
    time = "time"
    space = "space"


SpaceUnits = Literal[
    "micrometer",
    "nanometer",
    "angstrom",
    "picometer",
    "millimeter",
    "centimeter",
    "decimeter",
    "meter",
    "inch",
    "foot",
    "yard",
    "mile",
    "kilometer",
    "hectometer",
    "megameter",
    "gigameter",
    "terameter",
    "petameter",
    "exameter",
    "parsec",
    "femtometer",
    "attometer",
    "zeptometer",
    "yoctometer",
    "zettameter",
    "yottameter",
]
DefaultSpaceUnit = "micrometer"

TimeUnits = Literal[
    "attosecond",
    "centisecond",
    "day",
    "decisecond",
    "exasecond",
    "femtosecond",
    "gigasecond",
    "hectosecond",
    "hour",
    "kilosecond",
    "megasecond",
    "microsecond",
    "millisecond",
    "minute",
    "nanosecond",
    "petasecond",
    "picosecond",
    "second",
    "terasecond",
    "yoctosecond",
    "yottasecond",
    "zeptosecond",
    "zettasecond",
]
DefaultTimeUnit = "second"


class Axis(BaseModel):
    """Axis infos model."""

    name: str
    unit: str | None = None
    axis_type: AxisType | None = None

    model_config = ConfigDict(extra="forbid", frozen=True)

    def implicit_type_cast(self, cast_type: AxisType) -> "Axis":
        unit = self.unit
        if cast_type == AxisType.time and unit is None:
            unit = DefaultTimeUnit

        if cast_type == AxisType.space and unit is None:
            unit = DefaultSpaceUnit

        return Axis(name=self.name, axis_type=cast_type, unit=unit)

    def canonical_axis_cast(self, canonical_name: str) -> "Axis":
        """Cast the implicit axis to the correct type."""
        match canonical_name:
            case "t":
                if self.axis_type != AxisType.time or self.unit is None:
                    return self.implicit_type_cast(AxisType.time)
            case "c":
                if self.axis_type != AxisType.channel:
                    return self.implicit_type_cast(AxisType.channel)
            case "z" | "y" | "x":
                if self.axis_type != AxisType.space or self.unit is None:
                    return self.implicit_type_cast(AxisType.space)
        return self


################################################################################################
#
# Axes Handling
# We define a unique mapping to match the axes on disk to the canonical axes.
# The canonical axes are the ones that are used consistently in the NGIO internal API.
# The canonical axes ordered are: t, c, z, y, x.
#
#################################################################################################


def canonical_axes_order() -> tuple[str, str, str, str, str]:
    """Get the canonical axes order."""
    return "t", "c", "z", "y", "x"


def canonical_label_axes_order() -> tuple[str, str, str, str]:
    """Get the canonical axes order."""
    return "t", "z", "y", "x"


class AxesSetup(BaseModel):
    """Axes setup model.

    This model is used to map the on disk axes to the canonical OME-Zarr axes.
    """

    x: str = "x"
    y: str = "y"
    z: str = "z"
    c: str = "c"
    t: str = "t"
    others: list[str] = Field(default_factory=list)

    model_config = ConfigDict(extra="forbid", frozen=True)

    def canonical_map(self) -> dict[str, str]:
        """Get the canonical map of axes."""
        return {
            "t": self.t,
            "c": self.c,
            "z": self.z,
            "y": self.y,
            "x": self.x,
        }

    def get_on_disk_name(self, canonical_name: str) -> str | None:
        """Get the on disk name of the axis by its canonical name."""
        canonical_map = self.canonical_map()
        return canonical_map.get(canonical_name, None)

    def inverse_canonical_map(self) -> dict[str, str]:
        """Get the on disk map of axes."""
        return {
            self.t: "t",
            self.c: "c",
            self.z: "z",
            self.y: "y",
            self.x: "x",
        }

    def get_canonical_name(self, on_disk_name: str) -> str | None:
        """Get the canonical name of the axis by its on disk name."""
        inv_map = self.inverse_canonical_map()
        return inv_map.get(on_disk_name, None)


def _check_unique_names(axes: Sequence[Axis]):
    """Check if all axes on disk have unique names."""
    names = [ax.name for ax in axes]
    if len(set(names)) != len(names):
        duplicates = {item for item in names if names.count(item) > 1}
        raise NgioValidationError(
            f"All axes must be unique. But found duplicates axes {duplicates}"
        )


def _check_non_canonical_axes(axes_setup: AxesSetup, allow_non_canonical_axes: bool):
    """Check if all axes are known."""
    if not allow_non_canonical_axes and len(axes_setup.others) > 0:
        raise NgioValidationError(
            f"Unknown axes {axes_setup.others}. Please set "
            "`allow_non_canonical_axes=True` to ignore them"
        )


def _check_axes_validity(axes: Sequence[Axis], axes_setup: AxesSetup):
    """Check if all axes are valid."""
    _axes_setup = axes_setup.model_dump(exclude={"others"})
    _all_known_axes = [*_axes_setup.values(), *axes_setup.others]
    for ax in axes:
        if ax.name not in _all_known_axes:
            raise NgioValidationError(
                f"Invalid axis name '{ax.name}'. "
                f"Please correct map `{ax.name}` "
                f"using the AxesSetup model {axes_setup}"
            )


def _check_canonical_order(
    axes: Sequence[Axis], axes_setup: AxesSetup, strict_canonical_order: bool
):
    """Check if the axes are in the canonical order."""
    if not strict_canonical_order:
        return
    _names = [ax.name for ax in axes]
    _canonical_order = []
    for name in canonical_axes_order():
        mapped_name = getattr(axes_setup, name)
        if mapped_name in _names:
            _canonical_order.append(mapped_name)

    if _names != _canonical_order:
        raise NgioValidationError(
            f"Invalid axes order. The axes must be in the canonical order. "
            f"Expected {_canonical_order}, but found {_names}"
        )


def validate_axes(
    axes: Sequence[Axis],
    axes_setup: AxesSetup,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = False,
) -> None:
    """Validate the axes."""
    if allow_non_canonical_axes and strict_canonical_order:
        raise NgioValidationError(
            "`allow_non_canonical_axes` and"
            "`strict_canonical_order` cannot be true at the same time."
            "If non canonical axes are allowed, the order cannot be checked."
        )
    _check_unique_names(axes=axes)
    _check_non_canonical_axes(
        axes_setup=axes_setup, allow_non_canonical_axes=allow_non_canonical_axes
    )
    _check_axes_validity(axes=axes, axes_setup=axes_setup)
    _check_canonical_order(
        axes=axes, axes_setup=axes_setup, strict_canonical_order=strict_canonical_order
    )


class AxesHandler:
    """This class is used to handle and operate on OME-Zarr axes.

    The class also provides:
        - methods to reorder, squeeze and expand axes.
        - methods to validate the axes.
        - methods to get axis by name or index.
        - methods to operate on the axes.
    """

    def __init__(
        self,
        # spec dictated args
        axes: Sequence[Axis],
        # user defined args
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = False,
    ):
        """Create a new AxesMapper object.

        Args:
            axes (list[Axis]): The axes on disk.
            axes_setup (AxesSetup, optional): The axis setup. Defaults to None.
            allow_non_canonical_axes (bool, optional): Allow non canonical axes.
            strict_canonical_order (bool, optional): Check if the axes are in the
                canonical order. Defaults to False.
        """
        axes_setup = axes_setup if axes_setup is not None else AxesSetup()

        validate_axes(
            axes=axes,
            axes_setup=axes_setup,
            allow_non_canonical_axes=allow_non_canonical_axes,
            strict_canonical_order=strict_canonical_order,
        )

        self._allow_non_canonical_axes = allow_non_canonical_axes
        self._strict_canonical_order = strict_canonical_order

        self._canonical_order = canonical_axes_order()

        self._axes = axes
        self._axes_setup = axes_setup

        self._index_mapping = self._compute_index_mapping()

        # Validate the axes type and cast them if necessary
        # This needs to be done after the name mapping is computed
        self.validate_axes_type()

    def _compute_index_mapping(self):
        """Compute the index mapping.

        The index mapping is a dictionary with keys the canonical axes names
        and values the on disk axes index.

        Example:
            If the on disk axes are ['channel', 't', 'z', 'y', 'x'],
            the index mapping will be:
            {
                'c': 0,
                'channel': 0,
                't': 1,
                'z': 2,
                'y': 3,
                'x': 4,
            }
        """
        _index_mapping = {}
        for i, ax in enumerate(self.axes_names):
            _index_mapping[ax] = i
        # If the axis is not in the canonical order we also set it.
        canonical_map = self._axes_setup.canonical_map()
        for canonical_name, on_disk_name in canonical_map.items():
            if on_disk_name in _index_mapping.keys():
                _index_mapping[canonical_name] = _index_mapping[on_disk_name]
        return _index_mapping

    @property
    def axes_setup(self) -> AxesSetup:
        """Return the axes setup."""
        return self._axes_setup

    @property
    def axes(self) -> tuple[Axis, ...]:
        return tuple(self._axes)

    @property
    def axes_names(self) -> tuple[str, ...]:
        return tuple(ax.name for ax in self._axes)

    @property
    def allow_non_canonical_axes(self) -> bool:
        """Return if non canonical axes are allowed."""
        return self._allow_non_canonical_axes

    @property
    def strict_canonical_order(self) -> bool:
        """Return if strict canonical order is enforced."""
        return self._strict_canonical_order

    @property
    def space_unit(self) -> str | None:
        """Return the space unit for a given axis."""
        x_axis = self.get_axis("x")
        y_axis = self.get_axis("y")

        if x_axis is None or y_axis is None:
            raise NgioValidationError(
                "The dataset must have x and y axes to determine the space unit."
            )

        if x_axis.unit == y_axis.unit:
            return x_axis.unit
        else:
            raise NgioValidationError(
                "Inconsistent space units. "
                f"x={x_axis.unit} and y={y_axis.unit} should have the same unit."
            )

    @property
    def time_unit(self) -> str | None:
        """Return the time unit for a given axis."""
        t_axis = self.get_axis("t")
        if t_axis is None:
            return None
        return t_axis.unit

    def to_units(
        self,
        *,
        space_unit: SpaceUnits = DefaultSpaceUnit,
        time_unit: TimeUnits = DefaultTimeUnit,
    ) -> "AxesHandler":
        """Convert the pixel size to the given units.

        Args:
            space_unit(str): The space unit to convert to.
            time_unit(str): The time unit to convert to.
        """
        new_axes = []
        for ax in self.axes:
            if ax.axis_type == AxisType.space:
                new_ax = Axis(
                    name=ax.name,
                    axis_type=ax.axis_type,
                    unit=space_unit,
                )
                new_axes.append(new_ax)
            elif ax.axis_type == AxisType.time:
                new_ax = Axis(name=ax.name, axis_type=ax.axis_type, unit=time_unit)
                new_axes.append(new_ax)
            else:
                new_axes.append(ax)

        return AxesHandler(
            axes=new_axes,
            axes_setup=self.axes_setup,
            allow_non_canonical_axes=self.allow_non_canonical_axes,
            strict_canonical_order=self.strict_canonical_order,
        )

    def get_index(self, name: str) -> int | None:
        """Get the index of the axis by name."""
        return self._index_mapping.get(name, None)

    def has_axis(self, axis_name: str) -> bool:
        """Return whether the axis exists."""
        index = self.get_index(axis_name)
        if index is None:
            return False
        return True

    def get_canonical_name(self, name: str) -> str | None:
        """Get the canonical name of the axis by name."""
        return self._axes_setup.get_canonical_name(name)

    def get_axis(self, name: str) -> Axis | None:
        """Get the axis object by name."""
        index = self.get_index(name)
        if index is None:
            return None
        return self.axes[index]

    def validate_axes_type(self):
        """Validate the axes type.

        If the axes type is not correct, a warning is issued.
        and the axis is implicitly cast to the correct type.
        """
        new_axes = []
        for axes in self.axes:
            for name in self._canonical_order:
                if axes == self.get_axis(name):
                    new_axes.append(axes.canonical_axis_cast(name))
                    break
            else:
                new_axes.append(axes)
        self._axes = new_axes


def build_canonical_axes_handler(
    axes_names: Sequence[str],
    space_units: SpaceUnits | str | None = DefaultSpaceUnit,
    time_units: TimeUnits | str | None = DefaultTimeUnit,
    # user defined args
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = False,
) -> AxesHandler:
    """Create a new canonical axes mapper.

    Args:
        axes_names (Sequence[str] | int): The axes names on disk.
            - The axes should be in ['t', 'c', 'z', 'y', 'x']
            - The axes should be in strict canonical order.
            - If an integer is provided, the axes are created from the last axis
              to the first
                e.g. 3 -> ["z", "y", "x"]
        space_units (SpaceUnits, optional): The space units. Defaults to None.
        time_units (TimeUnits, optional): The time units. Defaults to None.
        axes_setup (AxesSetup, optional): The axis setup. Defaults to None.
        allow_non_canonical_axes (bool, optional): Allow non canonical axes.
            Defaults to False.
        strict_canonical_order (bool, optional): Check if the axes are in the
            canonical order. Defaults to False.

    """
    axes = []
    for name in axes_names:
        match name:
            case "t":
                axes.append(Axis(name=name, axis_type=AxisType.time, unit=time_units))
            case "c":
                axes.append(Axis(name=name, axis_type=AxisType.channel))
            case "z" | "y" | "x":
                axes.append(Axis(name=name, axis_type=AxisType.space, unit=space_units))
            case _:
                raise NgioValueError(
                    f"Invalid axis name '{name}'. "
                    "Only 't', 'c', 'z', 'y', 'x' are allowed."
                )

    return AxesHandler(
        axes=axes,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )



from collections.abc import Sequence
from typing import TypeAlias

import dask.array as da
import numpy as np

from ngio.common import Roi, RoiPixels
from ngio.experimental.iterators._abstract_iterator import AbstractIteratorBuilder
from ngio.images import Image, Label
from ngio.images._image import (
    ChannelSlicingInputType,
    add_channel_selection_to_slicing_dict,
)
from ngio.io_pipes import (
    DaskRoiGetter,
    DataGetter,
    NumpyRoiGetter,
    TransformProtocol,
)

NumpyPipeType: TypeAlias = tuple[np.ndarray, np.ndarray, Roi | RoiPixels]
DaskPipeType: TypeAlias = tuple[da.Array, da.Array, Roi | RoiPixels]


class NumpyFeatureGetter(DataGetter[NumpyPipeType]):
    def __init__(
        self,
        image_getter: NumpyRoiGetter,
        label_getter: NumpyRoiGetter,
    ) -> None:
        self._image_getter = image_getter
        self._label_getter = label_getter
        super().__init__(
            zarr_array=self._image_getter.zarr_array,
            slicing_ops=self._image_getter.slicing_ops,
            axes_ops=self._image_getter.axes_ops,
            transforms=self._image_getter.transforms,
            roi=self._image_getter.roi,
        )

    def get(self) -> NumpyPipeType:
        return self._image_getter(), self._label_getter(), self.roi

    @property
    def image(self) -> np.ndarray:
        return self._image_getter()

    @property
    def label(self) -> np.ndarray:
        return self._label_getter()


class DaskFeatureGetter(DataGetter[DaskPipeType]):
    def __init__(
        self,
        image_getter: DaskRoiGetter,
        label_getter: DaskRoiGetter,
    ) -> None:
        self._image_getter = image_getter
        self._label_getter = label_getter
        super().__init__(
            zarr_array=self._image_getter.zarr_array,
            slicing_ops=self._image_getter.slicing_ops,
            axes_ops=self._image_getter.axes_ops,
            transforms=self._image_getter.transforms,
            roi=self._image_getter.roi,
        )

    def get(self) -> DaskPipeType:
        return self._image_getter(), self._label_getter(), self.roi

    @property
    def image(self) -> da.Array:
        return self._image_getter()

    @property
    def label(self) -> da.Array:
        return self._label_getter()


class FeatureExtractorIterator(AbstractIteratorBuilder[NumpyPipeType, DaskPipeType]):
    """Base class for iterators over ROIs."""

    def __init__(
        self,
        input_image: Image,
        input_label: Label,
        channel_selection: ChannelSlicingInputType = None,
        axes_order: Sequence[str] | None = None,
        input_transforms: Sequence[TransformProtocol] | None = None,
        label_transforms: Sequence[TransformProtocol] | None = None,
    ) -> None:
        """Initialize the iterator with a ROI table and input/output images.

        Args:
            input_image (Image): The input image to be used as input for the
                segmentation.
            input_label (Label): The input label with the segmentation masks.
            channel_selection (ChannelSlicingInputType): Optional
                selection of channels to use for the segmentation.
            axes_order (Sequence[str] | None): Optional axes order for the
                segmentation.
            input_transforms (Sequence[TransformProtocol] | None): Optional
                transforms to apply to the input image.
            label_transforms (Sequence[TransformProtocol] | None): Optional
                transforms to apply to the output label.
        """
        self._input = input_image
        self._input_label = input_label
        self._ref_image = input_image
        self._rois = input_image.build_image_roi_table(name=None).rois()

        # Set iteration parameters
        self._input_slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self._input, channel_selection=channel_selection, slicing_dict={}
        )
        self._channel_selection = channel_selection
        self._axes_order = axes_order
        self._input_transforms = input_transforms
        self._label_transforms = label_transforms

        self._input.require_axes_match(self._input_label)
        self._input.require_rescalable(self._input_label)

    def get_init_kwargs(self) -> dict:
        """Return the initialization arguments for the iterator."""
        return {
            "input_image": self._input,
            "input_label": self._input_label,
            "channel_selection": self._channel_selection,
            "axes_order": self._axes_order,
            "input_transforms": self._input_transforms,
            "label_transforms": self._label_transforms,
        }

    def build_numpy_getter(self, roi: Roi) -> NumpyFeatureGetter:
        data_getter = NumpyRoiGetter(
            zarr_array=self._input.zarr_array,
            dimensions=self._input.dimensions,
            axes_order=self._axes_order,
            transforms=self._input_transforms,
            roi=roi,
            slicing_dict=self._input_slicing_kwargs,
        )
        label_getter = NumpyRoiGetter(
            zarr_array=self._input_label.zarr_array,
            dimensions=self._input_label.dimensions,
            axes_order=self._axes_order,
            transforms=self._label_transforms,
            roi=roi,
            remove_channel_selection=True,
        )
        return NumpyFeatureGetter(data_getter, label_getter)

    def build_dask_getter(self, roi: Roi) -> DaskFeatureGetter:
        data_getter = DaskRoiGetter(
            zarr_array=self._input.zarr_array,
            dimensions=self._input.dimensions,
            axes_order=self._axes_order,
            transforms=self._input_transforms,
            roi=roi,
            slicing_dict=self._input_slicing_kwargs,
        )
        label_getter = DaskRoiGetter(
            zarr_array=self._input_label.zarr_array,
            dimensions=self._input_label.dimensions,
            axes_order=self._axes_order,
            transforms=self._label_transforms,
            roi=roi,
            remove_channel_selection=True,
        )
        return DaskFeatureGetter(data_getter, label_getter)

    def build_numpy_setter(self, roi: Roi) -> None:
        return None

    def build_dask_setter(self, roi: Roi) -> None:
        return None

    def post_consolidate(self):
        pass

    def iter_as_numpy(self):  # type: ignore[override]
        """Create an iterator over the pixels of the ROIs."""
        return self.iter(lazy=False, data_mode="numpy", iterator_mode="readonly")

    def iter_as_dask(self):  # type: ignore[override]
        """Create an iterator over the pixels of the ROIs."""
        return self.iter(lazy=False, data_mode="dask", iterator_mode="readonly")



import math
from collections.abc import Mapping, Sequence
from typing import TypeAlias, assert_never
from warnings import warn

import dask.array as da
import numpy as np
import zarr
from pydantic import BaseModel, ConfigDict

from ngio.common._dimensions import Dimensions
from ngio.io_pipes._ops_slices_utils import compute_slice_chunks
from ngio.ome_zarr_meta.ngio_specs import Axis
from ngio.utils import NgioValueError

SlicingInputType: TypeAlias = slice | Sequence[int] | int | None
SlicingType: TypeAlias = slice | list[int] | int

##############################################################
#
# "SlicingOps" model
#
##############################################################


def _int_boundary_check(value: int, shape: int) -> int:
    """Ensure that the integer value is within the boundaries of the array shape."""
    if value < 0 or value >= shape:
        raise NgioValueError(
            f"Invalid index {value}. Index is out of bounds for axis with size {shape}."
        )
    return value


def _slicing_tuple_boundary_check(
    slicing_tuple: tuple[SlicingType, ...],
    array_shape: tuple[int, ...],
) -> tuple[SlicingType, ...]:
    """Ensure that the slicing tuple is within the boundaries of the array shape.

    This function normalizes the slicing tuple to ensure that the selection
    is within the boundaries of the array shape.
    """
    if len(slicing_tuple) != len(array_shape):
        raise NgioValueError(
            f"Invalid slicing tuple {slicing_tuple}. "
            f"Length {len(slicing_tuple)} does not match array shape {array_shape}."
        )
    out_slicing_tuple = []
    for sl, sh in zip(slicing_tuple, array_shape, strict=True):
        if isinstance(sl, slice):
            start, stop, step = sl.start, sl.stop, sl.step
            if start is not None:
                start = math.floor(start)
                start = max(0, min(start, sh))
            if stop is not None:
                stop = math.ceil(stop)
                stop = max(0, min(stop, sh))
            out_slicing_tuple.append(slice(start, stop, step))
        elif isinstance(sl, int):
            _int_boundary_check(sl, shape=sh)
            out_slicing_tuple.append(sl)
        elif isinstance(sl, list):
            [_int_boundary_check(i, shape=sh) for i in sl]
            out_slicing_tuple.append(sl)
        else:
            assert_never(sl)

    return tuple(out_slicing_tuple)


class SlicingOps(BaseModel):
    """Class to hold slicing operations."""

    on_disk_axes: tuple[str, ...]
    on_disk_shape: tuple[int, ...]
    on_disk_chunks: tuple[int, ...]
    slicing_tuple: tuple[SlicingType, ...]
    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)

    @property
    def normalized_slicing_tuple(self) -> tuple[SlicingType, ...]:
        """Normalize the slicing tuple to be within the array shape boundaries."""
        return _slicing_tuple_boundary_check(
            slicing_tuple=self.slicing_tuple,
            array_shape=self.on_disk_shape,
        )

    @property
    def slice_axes(self) -> tuple[str, ...]:
        """The axes after slicing."""
        in_memory_axes = []
        for ax, sl in zip(self.on_disk_axes, self.slicing_tuple, strict=True):
            if isinstance(sl, int):
                continue
            in_memory_axes.append(ax)
        return tuple(in_memory_axes)

    def slice_chunks(self) -> set[tuple[int, ...]]:
        """The required to read or write the slice."""
        return compute_slice_chunks(
            shape=self.on_disk_shape,
            chunks=self.on_disk_chunks,
            slicing_tuple=self.normalized_slicing_tuple,
        )

    def get(self, ax_name: str, normalize: bool = False) -> SlicingType:
        """Get the slicing tuple."""
        slicing_tuple = (
            self.slicing_tuple if not normalize else self.normalized_slicing_tuple
        )
        if ax_name not in self.on_disk_axes:
            return slice(None)
        ax_index = self.on_disk_axes.index(ax_name)
        return slicing_tuple[ax_index]


def _check_list_in_slicing_tuple(
    slicing_tuple: tuple[SlicingType, ...],
) -> tuple[None, None] | tuple[int, list[int]]:
    """Check if there are any lists in the slicing tuple.

    Dask regions when setting data do not support non-contiguous
    selections natively.
    Ngio support a single list in the slicing tuple to allow non-contiguous
    selection (main use case: selecting multiple channels).
    """
    # Find if the is any list in the slicing tuple
    # If there is one we need to handle it differently
    list_in_slice = [(i, s) for i, s in enumerate(slicing_tuple) if isinstance(s, list)]
    if not list_in_slice:
        # No list in the slicing tuple
        return None, None

    if len(list_in_slice) > 1:
        raise NotImplementedError(
            "Slicing with multiple non-contiguous tuples/lists "
            "is not supported yet in Ngio. Use directly the "
            "zarr.Array api to get the correct array slice."
        )
    # Complex case, we have exactly one tuple in the slicing tuple
    ax, first_tuple = list_in_slice[0]
    if len(first_tuple) > 100:
        warn(
            "Performance warning: "
            "Non-contiguous slicing with a tuple/list with more than 100 elements is "
            "not natively supported by zarr. This is implemented by Ngio by performing "
            "multiple reads and stacking the result.",
            stacklevel=2,
        )
    return ax, first_tuple


##############################################################
#
# Slicing implementations
#
##############################################################


def get_slice_as_numpy(zarr_array: zarr.Array, slicing_ops: SlicingOps) -> np.ndarray:
    """Get a slice of a zarr array as a numpy array."""
    slicing_tuple = slicing_ops.normalized_slicing_tuple
    # Find if the is any tuple in the slicing tuple
    # If there is one we need to handle it differently
    return zarr_array[slicing_tuple]


def get_slice_as_dask(zarr_array: zarr.Array, slicing_ops: SlicingOps) -> da.Array:
    """Get a slice of a zarr array as a dask array."""
    da_array = da.from_zarr(zarr_array)
    slicing_tuple = slicing_ops.normalized_slicing_tuple
    return da_array[slicing_tuple]


def set_slice_as_numpy(
    zarr_array: zarr.Array,
    patch: np.ndarray,
    slicing_ops: SlicingOps,
) -> None:
    slice_tuple = slicing_ops.normalized_slicing_tuple
    zarr_array[slice_tuple] = patch


def handle_int_set_as_dask(
    patch: da.Array,
    slicing_tuple: tuple[SlicingType, ...],
) -> tuple[da.Array, tuple[SlicingType, ...]]:
    """Handle the case where the slicing tuple contains integers.

    In this case we need to expand the patch array to match the slicing tuple.
    """
    new_slicing_tuple = list(slicing_tuple)
    for i, sl in enumerate(slicing_tuple):
        if isinstance(sl, int):
            patch = da.expand_dims(patch, axis=i)
            new_slicing_tuple[i] = slice(sl, sl + 1)
    return patch, tuple(new_slicing_tuple)


def set_slice_as_dask(
    zarr_array: zarr.Array, patch: da.Array, slicing_ops: SlicingOps
) -> None:
    slice_tuple = slicing_ops.normalized_slicing_tuple
    ax, first_tuple = _check_list_in_slicing_tuple(slice_tuple)
    patch, slice_tuple = handle_int_set_as_dask(patch, slice_tuple)
    if ax is None:
        # Base case, no tuple in the slicing tuple
        # assert False
        da.to_zarr(arr=patch, url=zarr_array, region=slice_tuple)
        return

    # Complex case, we have exactly one tuple in the slicing tuple
    assert first_tuple is not None
    for i, idx in enumerate(first_tuple):
        _sub_slice = (*slice_tuple[:ax], slice(idx, idx + 1), *slice_tuple[ax + 1 :])
        sub_patch = da.take(patch, indices=i, axis=ax)
        sub_patch = da.expand_dims(sub_patch, axis=ax)
        da.to_zarr(arr=sub_patch, url=zarr_array, region=_sub_slice)


##############################################################
#
# Builder functions
#
##############################################################


def _try_to_slice(value: Sequence[int]) -> slice | list[int]:
    """Try to convert a list of integers into a slice if they are contiguous.

    - If the input is empty, return an empty tuple.
    - If the input is sorted, and contains contiguous integers,
      return a slice from the minimum to the maximum integer.
    - Otherwise, return the input as a list of integers.

    This is useful for optimizing array slicing operations
    by allowing the use of slices when possible, which can be more efficient.
    """
    if not value:
        raise NgioValueError("Ngio does not support empty sequences as slice input.")

    if not all(isinstance(i, int) for i in value):
        _value = []
        for i in value:
            try:
                _value.append(int(i))
            except Exception as e:
                raise NgioValueError(
                    f"Invalid value {i} of type {type(i)} in sequence {value}"
                ) from e
        value = _value
    # If the input is not sorted, return it as a tuple
    max_input = max(value)
    min_input = min(value)
    assert min_input >= 0, "Input must contain non-negative integers"

    if sorted(value) == list(range(min_input, max_input + 1)):
        return slice(min_input, max_input + 1)

    return list(value)


def _remove_channel_slicing(
    slicing_dict: dict[str, SlicingInputType],
    dimensions: Dimensions,
) -> dict[str, SlicingInputType]:
    """This utility function removes the channel selection from the slice kwargs.

    if ignore_channel_selection is True, it will remove the channel selection
    regardless of the dimensions. If the ignore_channel_selection is False
    it will fail.
    """
    if dimensions.is_multi_channels:
        return slicing_dict

    if "c" in slicing_dict:
        slicing_dict.pop("c", None)
    return slicing_dict


def _check_slicing_virtual_axes(slice_: SlicingInputType) -> bool:
    """Check if the slice_ is compatible with virtual axes.

    Virtual axes are axes that are not present in the actual data,
    such as time or channel axes in some datasets.
    So the only valid slices for virtual axes are:
    - None: means all data along the axis
    - 0: means the first element along the axis
    - slice([0, None], [1, None])
    """
    if slice_ is None or slice_ == 0:
        return True
    if isinstance(slice_, slice):
        if slice_.start is None and slice_.stop is None:
            return True
        if slice_.start == 0 and slice_.stop is None:
            return True
        if slice_.start is None and slice_.stop == 0:
            return True
        if slice_.start == 0 and slice_.stop == 1:
            return True
    if isinstance(slice_, Sequence):
        if len(slice_) == 1 and slice_[0] == 0:
            return True
    return False


def _clean_slicing_dict(
    dimensions: Dimensions,
    slicing_dict: Mapping[str, SlicingInputType],
    remove_channel_selection: bool = False,
) -> dict[str, SlicingInputType]:
    """Clean the slicing dict.

    This function will:
        - Validate that the axes in the slicing_dict are present in the dimensions.
        - Make sure that the slicing_dict uses the on-disk axis names.
        - Check for duplicate axis names in the slicing_dict.
        - Clean up channel selection if the dimensions
    """
    clean_slicing_dict: dict[str, SlicingInputType] = {}
    for axis_name, slice_ in slicing_dict.items():
        axis = dimensions.axes_handler.get_axis(axis_name)
        if axis is None:
            # Virtual axes should be allowed to be selected
            # Common use case is still allowing channel_selection
            # When the zarr has not channel axis.
            if not _check_slicing_virtual_axes(slice_):
                raise NgioValueError(
                    f"Invalid axis selection:{axis_name}={slice_}. "
                    f"Not found on the on-disk axes {dimensions.axes}."
                )
            # Virtual axes can be safely ignored
            continue
        if axis.name in clean_slicing_dict:
            raise NgioValueError(
                f"Duplicate axis {axis.name} in slice kwargs. "
                "Please provide unique axis names."
            )
        clean_slicing_dict[axis.name] = slice_

    if remove_channel_selection:
        clean_slicing_dict = _remove_channel_slicing(
            slicing_dict=clean_slicing_dict, dimensions=dimensions
        )
    return clean_slicing_dict


def _normalize_slicing_tuple(
    axis: Axis,
    slicing_dict: dict[str, SlicingInputType],
) -> SlicingType:
    """Normalize the slicing dict to tuple.

    Since the slicing dict can contain different types of values
    We need to normalize them to more predictable types.
    The output types are:
    - slice
    - int
    - list of int (for non-contiguous selection)
    """
    axis_name = axis.name
    if axis_name not in slicing_dict:
        # If no slice is provided for the axis, use a full slice
        return slice(None)

    value = slicing_dict[axis_name]
    if value is None:
        return slice(None)
    if isinstance(value, slice) or isinstance(value, int):
        return value
    elif isinstance(value, Sequence):
        # If a contiguous sequence of integers is provided,
        # convert it to a slice for simplicity.
        # Alternatively, it will be converted to a list of ints
        return _try_to_slice(value)

    raise NgioValueError(
        f"Invalid slice definition {value} of type {type(value)}. "
        "Allowed types are: int, slice, sequence of int or None."
    )


def build_slicing_ops(
    *,
    dimensions: Dimensions,
    slicing_dict: dict[str, SlicingInputType] | None,
    remove_channel_selection: bool = False,
) -> SlicingOps:
    """Assemble slices to be used to query the array."""
    slicing_dict = slicing_dict or {}
    _slicing_dict = _clean_slicing_dict(
        dimensions=dimensions,
        slicing_dict=slicing_dict,
        remove_channel_selection=remove_channel_selection,
    )

    slicing_tuple = tuple(
        _normalize_slicing_tuple(
            axis=axis,
            slicing_dict=_slicing_dict,
        )
        for axis in dimensions.axes_handler.axes
    )
    return SlicingOps(
        on_disk_axes=dimensions.axes_handler.axes_names,
        on_disk_shape=dimensions.shape,
        on_disk_chunks=dimensions.chunks,
        slicing_tuple=slicing_tuple,
    )



"""OME-Zarr object models."""

from ngio.images._create_synt_container import create_synthetic_ome_zarr
from ngio.images._image import ChannelSelectionModel, Image, ImagesContainer
from ngio.images._label import Label, LabelsContainer
from ngio.images._ome_zarr_container import (
    OmeZarrContainer,
    create_empty_ome_zarr,
    create_ome_zarr_from_array,
    open_image,
    open_label,
    open_ome_zarr_container,
)
from ngio.images._table_ops import (
    concatenate_image_tables,
    concatenate_image_tables_as,
    concatenate_image_tables_as_async,
    concatenate_image_tables_async,
    conctatenate_tables,
    list_image_tables,
    list_image_tables_async,
)

__all__ = [
    "ChannelSelectionModel",
    "Image",
    "ImagesContainer",
    "Label",
    "LabelsContainer",
    "OmeZarrContainer",
    "concatenate_image_tables",
    "concatenate_image_tables_as",
    "concatenate_image_tables_as_async",
    "concatenate_image_tables_async",
    "conctatenate_tables",
    "create_empty_ome_zarr",
    "create_ome_zarr_from_array",
    "create_synthetic_ome_zarr",
    "list_image_tables",
    "list_image_tables_async",
    "open_image",
    "open_label",
    "open_ome_zarr_container",
]



"""A module for handling label images in OME-NGFF files."""

from collections.abc import Sequence
from typing import Literal

import dask.array as da
import numpy as np

from ngio.images._image import (
    ChannelSlicingInputType,
    Image,
    SlicingInputType,
    add_channel_selection_to_slicing_dict,
)
from ngio.images._label import Label
from ngio.io_pipes import (
    DaskGetterMasked,
    DaskSetterMasked,
    NumpyGetterMasked,
    NumpySetterMasked,
    TransformProtocol,
)
from ngio.ome_zarr_meta import ImageMetaHandler, LabelMetaHandler
from ngio.tables import MaskingRoiTable
from ngio.utils import (
    ZarrGroupHandler,
)


class MaskedImage(Image):
    """Placeholder class for a label."""

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        path: str,
        meta_handler: ImageMetaHandler | None,
        label: Label,
        masking_roi_table: MaskingRoiTable,
    ) -> None:
        """Initialize the Image at a single level.

        Args:
            group_handler: The Zarr group handler.
            path: The path to the image in the ome_zarr file.
            meta_handler: The image metadata handler.
            label: The label image.
            masking_roi_table: The masking ROI table.

        """
        super().__init__(
            group_handler=group_handler, path=path, meta_handler=meta_handler
        )
        self._label = label
        self._masking_roi_table = masking_roi_table

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        label_name = self._label.meta.name
        if label_name is None:
            label_name = self._masking_roi_table.reference_label
        return f"MaskedImage(path={self.path}, {self.dimensions}, {label_name})"

    def get_roi_as_numpy(  # type: ignore (this ignore the method override issue)
        self,
        label: int,
        channel_selection: ChannelSlicingInputType | None = None,
        zoom_factor: float = 1.0,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: slice | int | Sequence[int],
    ) -> np.ndarray:
        """Return the array for a given ROI."""
        roi = self._masking_roi_table.get_label(label)
        roi = roi.zoom(zoom_factor)
        return super().get_roi_as_numpy(
            roi=roi,
            channel_selection=channel_selection,
            axes_order=axes_order,
            transforms=transforms,
            **slicing_kwargs,
        )

    def get_roi_as_dask(  # type: ignore (this ignore the method override issue)
        self,
        label: int,
        channel_selection: ChannelSlicingInputType | None = None,
        zoom_factor: float = 1.0,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: slice | int | Sequence[int],
    ) -> da.Array:
        """Return the array for a given ROI as a Dask array."""
        roi = self._masking_roi_table.get_label(label)
        roi = roi.zoom(zoom_factor)
        return super().get_roi_as_dask(
            roi=roi,
            channel_selection=channel_selection,
            axes_order=axes_order,
            transforms=transforms,
            **slicing_kwargs,
        )

    def get_roi(  # type: ignore (this ignore the method override issue)
        self,
        label: int,
        zoom_factor: float = 1.0,
        channel_selection: ChannelSlicingInputType | None = None,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        mode: Literal["numpy", "dask"] = "numpy",
        **slicing_kwargs: slice | int | Sequence[int],
    ) -> np.ndarray | da.Array:
        """Return the array for a given ROI."""
        roi = self._masking_roi_table.get_label(label)
        roi = roi.zoom(zoom_factor)
        return super().get_roi(
            roi=roi,
            channel_selection=channel_selection,
            axes_order=axes_order,
            transforms=transforms,
            mode=mode,
            **slicing_kwargs,
        )

    def set_roi(  # type: ignore (this ignore the method override issue)
        self,
        label: int,
        patch: np.ndarray | da.Array,
        zoom_factor: float = 1.0,
        channel_selection: ChannelSlicingInputType | None = None,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: slice | int | Sequence[int],
    ) -> None:
        """Set the array for a given ROI."""
        roi = self._masking_roi_table.get_label(label)
        roi = roi.zoom(zoom_factor)
        return super().set_roi(
            roi=roi,
            patch=patch,
            channel_selection=channel_selection,
            axes_order=axes_order,
            transforms=transforms,
            **slicing_kwargs,
        )

    def get_roi_masked_as_numpy(
        self,
        label: int,
        channel_selection: ChannelSlicingInputType | None = None,
        zoom_factor: float = 1.0,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        allow_rescaling: bool = True,
        **slicing_kwargs: SlicingInputType,
    ) -> np.ndarray:
        """Return the masked array for a given label as a NumPy array."""
        slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self, channel_selection=channel_selection, slicing_dict=slicing_kwargs
        )

        roi = self._masking_roi_table.get_label(label)
        roi = roi.zoom(zoom_factor)
        masked_getter = NumpyGetterMasked(
            roi=roi,
            zarr_array=self.zarr_array,
            label_zarr_array=self._label.zarr_array,
            dimensions=self.dimensions,
            label_dimensions=self._label.dimensions,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=slicing_kwargs,
            allow_rescaling=allow_rescaling,
        )
        return masked_getter()

    def get_roi_masked_as_dask(
        self,
        label: int,
        channel_selection: ChannelSlicingInputType | None = None,
        zoom_factor: float = 1.0,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        allow_rescaling: bool = True,
        **slicing_kwargs: SlicingInputType,
    ) -> da.Array:
        """Return the masked array for a given label as a Dask array."""
        slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self, channel_selection=channel_selection, slicing_dict=slicing_kwargs
        )

        roi = self._masking_roi_table.get_label(label)
        roi = roi.zoom(zoom_factor)
        masked_getter = DaskGetterMasked(
            roi=roi,
            zarr_array=self.zarr_array,
            label_zarr_array=self._label.zarr_array,
            dimensions=self.dimensions,
            label_dimensions=self._label.dimensions,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=slicing_kwargs,
            allow_rescaling=allow_rescaling,
        )
        return masked_getter()

    def get_roi_masked(
        self,
        label: int,
        channel_selection: ChannelSlicingInputType | None = None,
        zoom_factor: float = 1.0,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        mode: Literal["numpy", "dask"] = "numpy",
        allow_rescaling: bool = True,
        **slicing_kwargs: SlicingInputType,
    ) -> np.ndarray | da.Array:
        """Return the masked array for a given label."""
        if mode == "numpy":
            return self.get_roi_masked_as_numpy(
                label=label,
                channel_selection=channel_selection,
                zoom_factor=zoom_factor,
                axes_order=axes_order,
                transforms=transforms,
                allow_rescaling=allow_rescaling,
                **slicing_kwargs,
            )

        elif mode == "dask":
            return self.get_roi_masked_as_dask(
                label=label,
                channel_selection=channel_selection,
                zoom_factor=zoom_factor,
                axes_order=axes_order,
                transforms=transforms,
                allow_rescaling=allow_rescaling,
                **slicing_kwargs,
            )
        else:
            raise ValueError(f"Unknown mode: {mode}")

    def set_roi_masked(
        self,
        label: int,
        patch: np.ndarray | da.Array,
        channel_selection: ChannelSlicingInputType | None = None,
        axes_order: Sequence[str] | None = None,
        zoom_factor: float = 1.0,
        transforms: Sequence[TransformProtocol] | None = None,
        allow_rescaling: bool = True,
        **slicing_kwargs: SlicingInputType,
    ) -> None:
        """Set the masked array for a given label."""
        slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self, channel_selection=channel_selection, slicing_dict=slicing_kwargs
        )

        roi = self._masking_roi_table.get_label(label)
        roi = roi.zoom(zoom_factor)
        if isinstance(patch, da.Array):
            path_setter = DaskSetterMasked(
                roi=roi,
                zarr_array=self.zarr_array,
                label_zarr_array=self._label.zarr_array,
                dimensions=self.dimensions,
                label_dimensions=self._label.dimensions,
                axes_order=axes_order,
                transforms=transforms,
                slicing_dict=slicing_kwargs,
                allow_rescaling=allow_rescaling,
            )
            path_setter(patch)
        elif isinstance(patch, np.ndarray):
            path_setter = NumpySetterMasked(
                roi=roi,
                zarr_array=self.zarr_array,
                label_zarr_array=self._label.zarr_array,
                dimensions=self.dimensions,
                label_dimensions=self._label.dimensions,
                axes_order=axes_order,
                transforms=transforms,
                slicing_dict=slicing_kwargs,
                allow_rescaling=allow_rescaling,
            )
            path_setter(patch)
        else:
            raise TypeError(
                f"Unsupported patch type: {type(patch)}. "
                "Expected numpy.ndarray or dask.array.Array."
            )


class MaskedLabel(Label):
    """Placeholder class for a label."""

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        path: str,
        meta_handler: LabelMetaHandler | None,
        label: Label,
        masking_roi_table: MaskingRoiTable,
    ) -> None:
        """Initialize the Image at a single level.

        Args:
            group_handler: The Zarr group handler.
            path: The path to the image in the ome_zarr file.
            meta_handler: The image metadata handler.
            label: The label image.
            masking_roi_table: The masking ROI table.

        """
        super().__init__(
            group_handler=group_handler, path=path, meta_handler=meta_handler
        )
        self._label = label
        self._masking_roi_table = masking_roi_table

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        label_name = self._label.meta.name
        if label_name is None:
            label_name = self._masking_roi_table.reference_label
        return f"MaskedLabel(path={self.path}, {self.dimensions}, {label_name})"

    def get_roi_as_numpy(
        self,
        label: int,
        zoom_factor: float = 1.0,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: slice | int | Sequence[int],
    ) -> np.ndarray:
        """Return the ROI as a NumPy array."""
        roi = self._masking_roi_table.get_label(label)
        roi = roi.zoom(zoom_factor)
        return super().get_roi_as_numpy(
            roi=roi,
            axes_order=axes_order,
            transforms=transforms,
            **slicing_kwargs,
        )

    def get_roi_as_dask(
        self,
        label: int,
        zoom_factor: float = 1.0,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: slice | int | Sequence[int],
    ) -> da.Array:
        """Return the ROI as a Dask array."""
        roi = self._masking_roi_table.get_label(label)
        roi = roi.zoom(zoom_factor)
        return super().get_roi_as_dask(
            roi=roi,
            axes_order=axes_order,
            transforms=transforms,
            **slicing_kwargs,
        )

    def get_roi(
        self,
        label: int,
        zoom_factor: float = 1.0,
        axes_order: Sequence[str] | None = None,
        mode: Literal["numpy", "dask"] = "numpy",
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: slice | int | Sequence[int],
    ) -> np.ndarray | da.Array:
        """Return the array for a given ROI."""
        roi = self._masking_roi_table.get_label(label)
        roi = roi.zoom(zoom_factor)
        return super().get_roi(
            roi=roi,
            axes_order=axes_order,
            mode=mode,
            transforms=transforms,
            **slicing_kwargs,
        )

    def set_roi(
        self,
        label: int,
        patch: np.ndarray | da.Array,
        zoom_factor: float = 1.0,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: slice | int | Sequence[int],
    ) -> None:
        """Set the array for a given ROI."""
        roi = self._masking_roi_table.get_label(label)
        roi = roi.zoom(zoom_factor)
        return super().set_roi(
            roi=roi,
            patch=patch,
            axes_order=axes_order,
            transforms=transforms,
            **slicing_kwargs,
        )

    def get_roi_masked_as_numpy(
        self,
        label: int,
        zoom_factor: float = 1.0,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        allow_rescaling: bool = True,
        **slicing_kwargs: SlicingInputType,
    ) -> np.ndarray:
        """Return the masked array for a given label as a NumPy array."""
        roi = self._masking_roi_table.get_label(label)
        roi = roi.zoom(zoom_factor)
        masked_getter = NumpyGetterMasked(
            roi=roi,
            zarr_array=self.zarr_array,
            label_zarr_array=self._label.zarr_array,
            dimensions=self.dimensions,
            label_dimensions=self._label.dimensions,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=slicing_kwargs,
            allow_rescaling=allow_rescaling,
        )
        return masked_getter()

    def get_roi_masked_as_dask(
        self,
        label: int,
        zoom_factor: float = 1.0,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        allow_rescaling: bool = True,
        **slicing_kwargs: SlicingInputType,
    ) -> da.Array:
        """Return the masked array for a given label as a Dask array."""
        roi = self._masking_roi_table.get_label(label)
        roi = roi.zoom(zoom_factor)
        masked_getter = DaskGetterMasked(
            roi=roi,
            zarr_array=self.zarr_array,
            label_zarr_array=self._label.zarr_array,
            dimensions=self.dimensions,
            label_dimensions=self._label.dimensions,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=slicing_kwargs,
            allow_rescaling=allow_rescaling,
        )
        return masked_getter()

    def get_roi_masked(
        self,
        label: int,
        zoom_factor: float = 1.0,
        axes_order: Sequence[str] | None = None,
        mode: Literal["numpy", "dask"] = "numpy",
        transforms: Sequence[TransformProtocol] | None = None,
        allow_rescaling: bool = True,
        **slicing_kwargs: SlicingInputType,
    ) -> np.ndarray | da.Array:
        """Return the masked array for a given label."""
        if mode == "numpy":
            return self.get_roi_masked_as_numpy(
                label=label,
                zoom_factor=zoom_factor,
                axes_order=axes_order,
                transforms=transforms,
                allow_rescaling=allow_rescaling,
                **slicing_kwargs,
            )

        elif mode == "dask":
            return self.get_roi_masked_as_dask(
                label=label,
                zoom_factor=zoom_factor,
                axes_order=axes_order,
                transforms=transforms,
                allow_rescaling=allow_rescaling,
                **slicing_kwargs,
            )
        else:
            raise ValueError(f"Unknown mode: {mode}")

    def set_roi_masked(
        self,
        label: int,
        patch: np.ndarray | da.Array,
        axes_order: Sequence[str] | None = None,
        zoom_factor: float = 1.0,
        transforms: Sequence[TransformProtocol] | None = None,
        allow_rescaling: bool = True,
        **slicing_kwargs: SlicingInputType,
    ) -> None:
        """Set the masked array for a given label."""
        roi = self._masking_roi_table.get_label(label)
        roi = roi.zoom(zoom_factor)
        if isinstance(patch, da.Array):
            path_setter = DaskSetterMasked(
                roi=roi,
                zarr_array=self.zarr_array,
                label_zarr_array=self._label.zarr_array,
                dimensions=self.dimensions,
                label_dimensions=self._label.dimensions,
                axes_order=axes_order,
                transforms=transforms,
                slicing_dict=slicing_kwargs,
                allow_rescaling=allow_rescaling,
            )
            path_setter(patch)
        elif isinstance(patch, np.ndarray):
            path_setter = NumpySetterMasked(
                roi=roi,
                zarr_array=self.zarr_array,
                label_zarr_array=self._label.zarr_array,
                dimensions=self.dimensions,
                label_dimensions=self._label.dimensions,
                axes_order=axes_order,
                transforms=transforms,
                slicing_dict=slicing_kwargs,
                allow_rescaling=allow_rescaling,
            )
            path_setter(patch)
        else:
            raise TypeError(
                f"Unsupported patch type: {type(patch)}. "
                "Expected numpy.ndarray or dask.array.Array."
            )



import pandas as pd
import polars as pl

from ngio.tables.backends._non_zarr_backends import NonZarrBaseBackend


def write_lf_to_parquet(path: str, table: pl.DataFrame) -> None:
    """Write a polars DataFrame to a Parquet file."""
    # make categorical into string (for pandas compatibility)
    schema = table.collect_schema()

    categorical_columns = []
    for name, dtype in zip(schema.names(), schema.dtypes(), strict=True):
        if dtype == pl.Categorical:
            categorical_columns.append(name)

    for col in categorical_columns:
        table = table.with_columns(pl.col(col).cast(pl.Utf8))

    # write to parquet
    table.write_parquet(path)


def write_df_to_parquet(path: str, table: pd.DataFrame) -> None:
    """Write a pandas DataFrame to a Parquet file."""
    table.to_parquet(path, index=False)


class ParquetTableBackend(NonZarrBaseBackend):
    """A class to load and write small tables in Parquet format."""

    def __init__(
        self,
    ):
        """Initialize the ParquetTableBackend."""
        super().__init__(
            lf_reader=pl.scan_parquet,
            df_reader=pd.read_parquet,
            lf_writer=write_lf_to_parquet,
            df_writer=write_df_to_parquet,
            table_name="table.parquet",
        )

    @staticmethod
    def backend_name() -> str:
        """Return the name of the backend."""
        return "parquet"



"""Abstract class for handling OME-NGFF images."""

import warnings
from collections.abc import Sequence
from typing import Literal

import numpy as np
from zarr.core.array import CompressorLike

from ngio.images._create import create_empty_image_container
from ngio.images._image import Image, ImagesContainer
from ngio.images._label import Label, LabelsContainer
from ngio.images._masked_image import MaskedImage, MaskedLabel
from ngio.ome_zarr_meta import NgioImageMeta, PixelSize, find_label_meta_handler
from ngio.ome_zarr_meta.ngio_specs import (
    DefaultNgffVersion,
    DefaultSpaceUnit,
    DefaultTimeUnit,
    NgffVersions,
    SpaceUnits,
    TimeUnits,
)
from ngio.tables import (
    ConditionTable,
    DefaultTableBackend,
    FeatureTable,
    GenericRoiTable,
    MaskingRoiTable,
    RoiTable,
    Table,
    TableBackend,
    TablesContainer,
    TableType,
    TypedTable,
)
from ngio.utils import (
    AccessModeLiteral,
    NgioValidationError,
    NgioValueError,
    StoreOrGroup,
    ZarrGroupHandler,
)


def _default_table_container(handler: ZarrGroupHandler) -> TablesContainer | None:
    """Return a default table container."""
    success, table_handler = handler.safe_derive_handler("tables")
    if success and isinstance(table_handler, ZarrGroupHandler):
        return TablesContainer(table_handler)


def _default_label_container(handler: ZarrGroupHandler) -> LabelsContainer | None:
    """Return a default label container."""
    success, label_handler = handler.safe_derive_handler("labels")
    if success and isinstance(label_handler, ZarrGroupHandler):
        return LabelsContainer(label_handler)


class OmeZarrContainer:
    """This class is an object representation of an OME-Zarr image.

    It provides methods to access:
        - The multiscale image metadata
        - To open images at different levels of resolution
        - To access labels and tables associated with the image.
        - To derive new images, labels, and add tables to the image.
        - To modify the image metadata, such as axes units and channel metadata.

    Attributes:
        images_container (ImagesContainer): The container for the images.
        labels_container (LabelsContainer): The container for the labels.
        tables_container (TablesContainer): The container for the tables.

    """

    _images_container: ImagesContainer
    _labels_container: LabelsContainer | None
    _tables_container: TablesContainer | None

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        table_container: TablesContainer | None = None,
        label_container: LabelsContainer | None = None,
        validate_paths: bool = False,
    ) -> None:
        """Initialize the OmeZarrContainer.

        Args:
            group_handler (ZarrGroupHandler): The Zarr group handler.
            table_container (TablesContainer | None): The tables container.
            label_container (LabelsContainer | None): The labels container.
            validate_paths (bool): Whether to validate the paths of the image multiscale
        """
        self._group_handler = group_handler
        self._images_container = ImagesContainer(self._group_handler)

        self._labels_container = label_container
        self._tables_container = table_container

        if validate_paths:
            for level_path in self._images_container.levels_paths:
                self.get_image(path=level_path)

    def __repr__(self) -> str:
        """Return a string representation of the image."""
        num_labels = len(self.list_labels())
        num_tables = len(self.list_tables())

        base_str = f"OmeZarrContainer(levels={self.levels}"
        if num_labels > 0 and num_labels < 3:
            base_str += f", labels={self.list_labels()}"
        elif num_labels >= 3:
            base_str += f", #labels={num_labels}"
        if num_tables > 0 and num_tables < 3:
            base_str += f", tables={self.list_tables()}"
        elif num_tables >= 3:
            base_str += f", #tables={num_tables}"
        base_str += ")"
        return base_str

    @property
    def images_container(self) -> ImagesContainer:
        """Return the images container.

        Returns:
            ImagesContainer: The images container.
        """
        return self._images_container

    def _get_labels_container(self) -> LabelsContainer | None:
        """Return the labels container."""
        if self._labels_container is None:
            _labels_container = _default_label_container(self._group_handler)
            if _labels_container is None:
                return None
            self._labels_container = _labels_container
        return self._labels_container

    @property
    def labels_container(self) -> LabelsContainer:
        """Return the labels container."""
        _labels_container = self._get_labels_container()
        if _labels_container is None:
            raise NgioValidationError("No labels found in the image.")
        return _labels_container

    def _get_tables_container(self) -> TablesContainer | None:
        """Return the tables container."""
        if self._tables_container is None:
            _tables_container = _default_table_container(self._group_handler)
            if _tables_container is None:
                return None
            self._tables_container = _tables_container
        return self._tables_container

    @property
    def tables_container(self) -> TablesContainer:
        """Return the tables container."""
        _tables_container = self._get_tables_container()
        if _tables_container is None:
            raise NgioValidationError("No tables found in the image.")
        return _tables_container

    @property
    def image_meta(self) -> NgioImageMeta:
        """Return the image metadata."""
        return self._images_container.meta

    @property
    def levels(self) -> int:
        """Return the number of levels in the image."""
        return self._images_container.levels

    @property
    def levels_paths(self) -> list[str]:
        """Return the paths of the levels in the image."""
        return self._images_container.levels_paths

    @property
    def is_3d(self) -> bool:
        """Return True if the image is 3D."""
        return self.get_image().is_3d

    @property
    def is_2d(self) -> bool:
        """Return True if the image is 2D."""
        return self.get_image().is_2d

    @property
    def is_time_series(self) -> bool:
        """Return True if the image is a time series."""
        return self.get_image().is_time_series

    @property
    def is_2d_time_series(self) -> bool:
        """Return True if the image is a 2D time series."""
        return self.get_image().is_2d_time_series

    @property
    def is_3d_time_series(self) -> bool:
        """Return True if the image is a 3D time series."""
        return self.get_image().is_3d_time_series

    @property
    def is_multi_channels(self) -> bool:
        """Return True if the image is multichannel."""
        return self.get_image().is_multi_channels

    @property
    def space_unit(self) -> str | None:
        """Return the space unit of the image."""
        return self.image_meta.space_unit

    @property
    def time_unit(self) -> str | None:
        """Return the time unit of the image."""
        return self.image_meta.time_unit

    @property
    def channel_labels(self) -> list[str]:
        """Return the channels of the image."""
        image = self.get_image()
        return image.channel_labels

    @property
    def wavelength_ids(self) -> list[str | None]:
        """Return the list of wavelength of the image."""
        image = self.get_image()
        return image.wavelength_ids

    @property
    def num_channels(self) -> int:
        """Return the number of channels."""
        return len(self.channel_labels)

    def get_channel_idx(
        self, channel_label: str | None = None, wavelength_id: str | None = None
    ) -> int:
        """Get the index of a channel by its label or wavelength ID."""
        image = self.get_image()
        return image.channels_meta.get_channel_idx(
            channel_label=channel_label, wavelength_id=wavelength_id
        )

    def set_channel_meta(
        self,
        labels: Sequence[str] | int | None = None,
        wavelength_id: Sequence[str] | None = None,
        percentiles: tuple[float, float] | None = None,
        colors: Sequence[str] | None = None,
        active: Sequence[bool] | None = None,
        **omero_kwargs: dict,
    ) -> None:
        """Create a ChannelsMeta object with the default unit."""
        self._images_container.set_channel_meta(
            labels=labels,
            wavelength_id=wavelength_id,
            start=None,
            end=None,
            percentiles=percentiles,
            colors=colors,
            active=active,
            **omero_kwargs,
        )

    def set_channel_percentiles(
        self,
        start_percentile: float = 0.1,
        end_percentile: float = 99.9,
    ) -> None:
        """Update the percentiles of the image."""
        self._images_container.set_channel_percentiles(
            start_percentile=start_percentile, end_percentile=end_percentile
        )

    def set_axes_units(
        self,
        space_unit: SpaceUnits = DefaultSpaceUnit,
        time_unit: TimeUnits = DefaultTimeUnit,
        set_labels: bool = True,
    ) -> None:
        """Set the units of the image.

        Args:
            space_unit (SpaceUnits): The unit of space.
            time_unit (TimeUnits): The unit of time.
            set_labels (bool): Whether to set the units for the labels as well.
        """
        self._images_container.set_axes_unit(space_unit=space_unit, time_unit=time_unit)
        if not set_labels:
            return
        for label_name in self.list_labels():
            label = self.get_label(label_name)
            label.set_axes_unit(space_unit=space_unit, time_unit=time_unit)

    def get_image(
        self,
        path: str | None = None,
        pixel_size: PixelSize | None = None,
        strict: bool = False,
    ) -> Image:
        """Get an image at a specific level.

        Args:
            path (str | None): The path to the image in the ome_zarr file.
            pixel_size (PixelSize | None): The pixel size of the image.
            strict (bool): Only used if the pixel size is provided. If True, the
                pixel size must match the image pixel size exactly. If False, the
                closest pixel size level will be returned.

        """
        return self._images_container.get(
            path=path, pixel_size=pixel_size, strict=strict
        )

    def _find_matching_masking_label(
        self,
        masking_label_name: str | None = None,
        masking_table_name: str | None = None,
        pixel_size: PixelSize | None = None,
    ) -> tuple[Label, MaskingRoiTable]:
        if masking_label_name is not None and masking_table_name is not None:
            # Both provided
            masking_label = self.get_label(
                name=masking_label_name, pixel_size=pixel_size, strict=False
            )
            masking_table = self.get_masking_roi_table(name=masking_table_name)

        elif masking_label_name is not None and masking_table_name is None:
            # Only the label provided
            masking_label = self.get_label(
                name=masking_label_name, pixel_size=pixel_size, strict=False
            )

            for table_name in self.list_roi_tables():
                table = self.get_generic_roi_table(name=table_name)
                if isinstance(table, MaskingRoiTable):
                    if table.reference_label == masking_label_name:
                        masking_table = table
                        break
            else:
                masking_table = masking_label.build_masking_roi_table()

        elif masking_table_name is not None and masking_label_name is None:
            # Only the table provided
            masking_table = self.get_masking_roi_table(name=masking_table_name)

            if masking_table.reference_label is None:
                raise NgioValueError(
                    f"Masking table {masking_table_name} does not have a reference "
                    "label. Please provide the masking_label_name explicitly."
                )
            masking_label = self.get_label(
                name=masking_table.reference_label,
                pixel_size=pixel_size,
                strict=False,
            )
        else:
            raise NgioValueError(
                "Neither masking_label_name nor masking_table_name were provided."
            )
        return masking_label, masking_table

    def get_masked_image(
        self,
        masking_label_name: str | None = None,
        masking_table_name: str | None = None,
        path: str | None = None,
        pixel_size: PixelSize | None = None,
        strict: bool = False,
    ) -> MaskedImage:
        """Get a masked image at a specific level.

        Args:
            masking_label_name (str | None): The name of the masking label to use.
                If None, the masking table must be provided.
            masking_table_name (str | None): The name of the masking table to use.
                If None, the masking label must be provided.
            path (str | None): The path to the image in the ome_zarr file.
                If None, the first level will be used.
            pixel_size (PixelSize | None): The pixel size of the image.
                This is only used if path is None.
            strict (bool): Only used if the pixel size is provided. If True, the
                pixel size must match the image pixel size exactly. If False, the
                closest pixel size level will be returned.
        """
        image = self.get_image(path=path, pixel_size=pixel_size, strict=strict)
        masking_label, masking_table = self._find_matching_masking_label(
            masking_label_name=masking_label_name,
            masking_table_name=masking_table_name,
            pixel_size=pixel_size,
        )
        return MaskedImage(
            group_handler=image._group_handler,
            path=image.path,
            meta_handler=image.meta_handler,
            label=masking_label,
            masking_roi_table=masking_table,
        )

    def derive_image(
        self,
        store: StoreOrGroup,
        ref_path: str | None = None,
        shape: Sequence[int] | None = None,
        labels: Sequence[str] | None = None,
        pixel_size: PixelSize | None = None,
        axes_names: Sequence[str] | None = None,
        name: str | None = None,
        chunks: Sequence[int] | None = None,
        dtype: str | None = None,
        dimension_separator: Literal[".", "/"] | None = None,
        compressors: CompressorLike | None = None,
        copy_labels: bool = False,
        copy_tables: bool = False,
        ngff_version: NgffVersions | None = None,
        overwrite: bool = False,
    ) -> "OmeZarrContainer":
        """Create an empty OME-Zarr container from an existing image.

        Args:
            store (StoreOrGroup): The Zarr store or group to create the image in.
            ref_path (str | None): The path to the reference image in
                the image container.
            shape (Sequence[int] | None): The shape of the new image.
            labels (Sequence[str] | None): The labels of the new image.
            pixel_size (PixelSize | None): The pixel size of the new image.
            axes_names (Sequence[str] | None): The axes names of the new image.
            chunks (Sequence[int] | None): The chunk shape of the new image.
            dtype (str | None): The data type of the new image.
            name (str | None): The name of the new image.
            dimension_separator (DIMENSION_SEPARATOR | None): The dimension
                separator to use. If None, the dimension separator of the
                reference image will be used.
            compressors (CompressorLike | None): The compressors to use. If None,
                the compressors of the reference image will be used.
            copy_labels (bool): Whether to copy the labels from the reference image.
            copy_tables (bool): Whether to copy the tables from the reference image.
            ngff_version (NgffVersions): The NGFF version to use.
            overwrite (bool): Whether to overwrite an existing image.

        Returns:
            OmeZarrContainer: The new image container.

        """
        new_container = self._images_container.derive(
            store=store,
            ref_path=ref_path,
            shape=shape,
            labels=labels,
            pixel_size=pixel_size,
            axes_names=axes_names,
            name=name,
            chunks=chunks,
            dtype=dtype,
            dimension_separator=dimension_separator,
            compressors=compressors,
            ngff_version=ngff_version,
            overwrite=overwrite,
        )

        new_ome_zarr = OmeZarrContainer(
            group_handler=new_container._group_handler,
            validate_paths=False,
        )

        if copy_labels:
            self.labels_container._group_handler.copy_handler(
                new_ome_zarr.labels_container._group_handler
            )

        if copy_tables:
            self.tables_container._group_handler.copy_handler(
                new_ome_zarr.tables_container._group_handler
            )
        return new_ome_zarr

    def list_tables(self, filter_types: TypedTable | str | None = None) -> list[str]:
        """List all tables in the image."""
        table_container = self._get_tables_container()
        if table_container is None:
            return []

        return table_container.list(
            filter_types=filter_types,
        )

    def list_roi_tables(self) -> list[str]:
        """List all ROI tables in the image."""
        masking_roi = self.tables_container.list(
            filter_types="masking_roi_table",
        )
        roi = self.tables_container.list(
            filter_types="roi_table",
        )
        return masking_roi + roi

    def get_roi_table(self, name: str) -> RoiTable:
        """Get a ROI table from the image.

        Args:
            name (str): The name of the table.
        """
        table = self.tables_container.get(name=name, strict=True)
        if not isinstance(table, RoiTable):
            raise NgioValueError(f"Table {name} is not a ROI table. Got {type(table)}")
        return table

    def get_masking_roi_table(self, name: str) -> MaskingRoiTable:
        """Get a masking ROI table from the image.

        Args:
            name (str): The name of the table.
        """
        table = self.tables_container.get(name=name, strict=True)
        if not isinstance(table, MaskingRoiTable):
            raise NgioValueError(
                f"Table {name} is not a masking ROI table. Got {type(table)}"
            )
        return table

    def get_feature_table(self, name: str) -> FeatureTable:
        """Get a feature table from the image.

        Args:
            name (str): The name of the table.
        """
        table = self.tables_container.get(name=name, strict=True)
        if not isinstance(table, FeatureTable):
            raise NgioValueError(
                f"Table {name} is not a feature table. Got {type(table)}"
            )
        return table

    def get_generic_roi_table(self, name: str) -> GenericRoiTable:
        """Get a generic ROI table from the image.

        Args:
            name (str): The name of the table.
        """
        table = self.tables_container.get(name=name, strict=True)
        if not isinstance(table, GenericRoiTable):
            raise NgioValueError(
                f"Table {name} is not a generic ROI table. Got {type(table)}"
            )
        return table

    def get_condition_table(self, name: str) -> ConditionTable:
        """Get a condition table from the image.

        Args:
            name (str): The name of the table.
        """
        table = self.tables_container.get(name=name, strict=True)
        if not isinstance(table, ConditionTable):
            raise NgioValueError(
                f"Table {name} is not a condition table. Got {type(table)}"
            )
        return table

    def get_table(self, name: str, check_type: TypedTable | None = None) -> Table:
        """Get a table from the image.

        Args:
            name (str): The name of the table.
            check_type (TypedTable | None): Deprecated. Please use
                'get_table_as' instead, or one of the type specific
                get_*table() methods.

        """
        if check_type is not None:
            warnings.warn(
                "The 'check_type' argument is deprecated, and will be removed in "
                "ngio=0.3. Use 'get_table_as' instead or one of the "
                "type specific get_*table() methods.",
                DeprecationWarning,
                stacklevel=2,
            )
        return self.tables_container.get(name=name, strict=False)

    def get_table_as(
        self,
        name: str,
        table_cls: type[TableType],
        backend: TableBackend | None = None,
    ) -> TableType:
        """Get a table from the image as a specific type.

        Args:
            name (str): The name of the table.
            table_cls (type[TableType]): The type of the table.
            backend (TableBackend | None): The backend to use. If None,
                the default backend is used.
        """
        return self.tables_container.get_as(
            name=name,
            table_cls=table_cls,
            backend=backend,
        )

    def build_image_roi_table(self, name: str | None = "image") -> RoiTable:
        """Compute the ROI table for an image."""
        return self.get_image().build_image_roi_table(name=name)

    def build_masking_roi_table(self, label: str) -> MaskingRoiTable:
        """Compute the masking ROI table for a label."""
        return self.get_label(label).build_masking_roi_table()

    def add_table(
        self,
        name: str,
        table: Table,
        backend: TableBackend = DefaultTableBackend,
        overwrite: bool = False,
    ) -> None:
        """Add a table to the image."""
        self.tables_container.add(
            name=name, table=table, backend=backend, overwrite=overwrite
        )

    def list_labels(self) -> list[str]:
        """List all labels in the image."""
        label_container = self._get_labels_container()
        if label_container is None:
            return []
        return label_container.list()

    def get_label(
        self,
        name: str,
        path: str | None = None,
        pixel_size: PixelSize | None = None,
        strict: bool = False,
    ) -> Label:
        """Get a label from the group.

        Args:
            name (str): The name of the label.
            path (str | None): The path to the image in the ome_zarr file.
            pixel_size (PixelSize | None): The pixel size of the image.
            strict (bool): Only used if the pixel size is provided. If True, the
                pixel size must match the image pixel size exactly. If False, the
                closest pixel size level will be returned.
        """
        return self.labels_container.get(
            name=name, path=path, pixel_size=pixel_size, strict=strict
        )

    def get_masked_label(
        self,
        label_name: str,
        masking_label_name: str | None = None,
        masking_table_name: str | None = None,
        path: str | None = None,
        pixel_size: PixelSize | None = None,
        strict: bool = False,
    ) -> MaskedLabel:
        """Get a masked image at a specific level.

        Args:
            label_name (str): The name of the label.
            masking_label_name (str | None): The name of the masking label.
            masking_table_name (str | None): The name of the masking table.
            path (str | None): The path to the image in the ome_zarr file.
            pixel_size (PixelSize | None): The pixel size of the image.
            strict (bool): Only used if the pixel size is provided. If True, the
                pixel size must match the image pixel size exactly. If False, the
                closest pixel size level will be returned.
        """
        label = self.get_label(
            name=label_name, path=path, pixel_size=pixel_size, strict=strict
        )
        masking_label, masking_table = self._find_matching_masking_label(
            masking_label_name=masking_label_name,
            masking_table_name=masking_table_name,
            pixel_size=pixel_size,
        )
        return MaskedLabel(
            group_handler=label._group_handler,
            path=label.path,
            meta_handler=label.meta_handler,
            label=masking_label,
            masking_roi_table=masking_table,
        )

    def derive_label(
        self,
        name: str,
        ref_image: Image | Label | None = None,
        shape: Sequence[int] | None = None,
        pixel_size: PixelSize | None = None,
        axes_names: Sequence[str] | None = None,
        chunks: Sequence[int] | None = None,
        dtype: str = "uint32",
        dimension_separator: Literal[".", "/"] | None = None,
        compressors: CompressorLike | None = None,
        overwrite: bool = False,
    ) -> "Label":
        """Create an empty OME-Zarr label from a reference image.

        And add the label to the /labels group.

        Args:
            name (str): The name of the new image.
            ref_image (Image | Label | None): A reference image that will be used
                to create the new image.
            shape (Sequence[int] | None): The shape of the new image.
            pixel_size (PixelSize | None): The pixel size of the new image.
            axes_names (Sequence[str] | None): The axes names of the new image.
                For labels, the channel axis is not allowed.
            chunks (Sequence[int] | None): The chunk shape of the new image.
            dtype (str): The data type of the new label.
            dimension_separator (DIMENSION_SEPARATOR | None): The dimension
                separator to use. If None, the dimension separator of the
                reference image will be used.
            compressors (CompressorLike | None): The compressors to use. If None,
                the compressors of the reference image will be used.
            overwrite (bool): Whether to overwrite an existing image.

        Returns:
            Label: The new label.

        """
        if ref_image is None:
            ref_image = self.get_image()
        return self.labels_container.derive(
            name=name,
            ref_image=ref_image,
            shape=shape,
            pixel_size=pixel_size,
            axes_names=axes_names,
            chunks=chunks,
            dtype=dtype,
            dimension_separator=dimension_separator,
            compressors=compressors,
            overwrite=overwrite,
        )


def open_ome_zarr_container(
    store: StoreOrGroup,
    cache: bool = False,
    mode: AccessModeLiteral = "r+",
    validate_arrays: bool = True,
) -> OmeZarrContainer:
    """Open an OME-Zarr image."""
    handler = ZarrGroupHandler(store=store, cache=cache, mode=mode)
    return OmeZarrContainer(
        group_handler=handler,
        validate_paths=validate_arrays,
    )


def open_image(
    store: StoreOrGroup,
    path: str | None = None,
    pixel_size: PixelSize | None = None,
    strict: bool = True,
    cache: bool = False,
    mode: AccessModeLiteral = "r+",
) -> Image:
    """Open a single level image from an OME-Zarr image.

    Args:
        store (StoreOrGroup): The Zarr store or group to create the image in.
        path (str | None): The path to the image in the ome_zarr file.
        pixel_size (PixelSize | None): The pixel size of the image.
        strict (bool): Only used if the pixel size is provided. If True, the
                pixel size must match the image pixel size exactly. If False, the
                closest pixel size level will be returned.
        cache (bool): Whether to use a cache for the zarr group metadata.
        mode (AccessModeLiteral): The
            access mode for the image. Defaults to "r+".
    """
    group_handler = ZarrGroupHandler(store=store, cache=cache, mode=mode)
    images_container = ImagesContainer(group_handler)
    return images_container.get(
        path=path,
        pixel_size=pixel_size,
        strict=strict,
    )


def open_label(
    store: StoreOrGroup,
    name: str | None = None,
    path: str | None = None,
    pixel_size: PixelSize | None = None,
    strict: bool = True,
    cache: bool = False,
    mode: AccessModeLiteral = "r+",
) -> Label:
    """Open a single level label from an OME-Zarr Label group.

    Args:
        store (StoreOrGroup): The Zarr store or group to create the image in.
        name (str | None): The name of the label. If None,
            we will try to open the store as a multiscale label.
        path (str | None): The path to the image in the ome_zarr file.
        pixel_size (PixelSize | None): The pixel size of the image.
        strict (bool): Only used if the pixel size is provided. If True, the
            pixel size must match the image pixel size exactly. If False, the
            closest pixel size level will be returned.
        cache (bool): Whether to use a cache for the zarr group metadata.
        mode (AccessModeLiteral): The access mode for the image. Defaults to "r+".

    """
    group_handler = ZarrGroupHandler(store=store, cache=cache, mode=mode)
    if name is None:
        label_meta_handler = find_label_meta_handler(group_handler)
        path = label_meta_handler.meta.get_dataset(
            path=path, pixel_size=pixel_size, strict=strict
        ).path
        return Label(group_handler, path, label_meta_handler)

    labels_container = LabelsContainer(group_handler)
    return labels_container.get(
        name=name,
        path=path,
        pixel_size=pixel_size,
        strict=strict,
    )


def create_empty_ome_zarr(
    store: StoreOrGroup,
    shape: Sequence[int],
    xy_pixelsize: float,
    z_spacing: float = 1.0,
    time_spacing: float = 1.0,
    levels: int | list[str] = 5,
    xy_scaling_factor: float = 2,
    z_scaling_factor: float = 1.0,
    space_unit: SpaceUnits = DefaultSpaceUnit,
    time_unit: TimeUnits = DefaultTimeUnit,
    axes_names: Sequence[str] | None = None,
    name: str | None = None,
    chunks: Sequence[int] | Literal["auto"] = "auto",
    dtype: str = "uint16",
    dimension_separator: Literal[".", "/"] = "/",
    compressors: CompressorLike = "auto",
    channel_labels: list[str] | None = None,
    channel_wavelengths: list[str] | None = None,
    channel_colors: Sequence[str] | None = None,
    channel_active: Sequence[bool] | None = None,
    overwrite: bool = False,
    version: NgffVersions = DefaultNgffVersion,
) -> OmeZarrContainer:
    """Create an empty OME-Zarr image with the given shape and metadata.

    Args:
        store (StoreOrGroup): The Zarr store or group to create the image in.
        shape (Sequence[int]): The shape of the image.
        xy_pixelsize (float): The pixel size in x and y dimensions.
        z_spacing (float, optional): The spacing between z slices. Defaults to 1.0.
        time_spacing (float, optional): The spacing between time points.
            Defaults to 1.0.
        levels (int | list[str], optional): The number of levels in the pyramid or a
            list of level names. Defaults to 5.
        xy_scaling_factor (float, optional): The down-scaling factor in x and y
            dimensions. Defaults to 2.0.
        z_scaling_factor (float, optional): The down-scaling factor in z dimension.
            Defaults to 1.0.
        space_unit (SpaceUnits, optional): The unit of space. Defaults to
            DefaultSpaceUnit.
        time_unit (TimeUnits, optional): The unit of time. Defaults to
            DefaultTimeUnit.
        axes_names (Sequence[str] | None, optional): The names of the axes.
            If None the canonical names are used. Defaults to None.
        name (str | None, optional): The name of the image. Defaults to None.
        chunks (Sequence[int] | None, optional): The chunk shape. If None the shape
            is used. Defaults to None.
        dtype (str, optional): The data type of the image. Defaults to "uint16".
        dimension_separator (DIMENSION_SEPARATOR): The dimension
            separator to use. Defaults to "/".
        compressors (CompressorLike): The compressor to use. Defaults to "auto".
        channel_labels (list[str] | None, optional): The labels of the channels.
            Defaults to None.
        channel_wavelengths (list[str] | None, optional): The wavelengths of the
            channels. Defaults to None.
        channel_colors (Sequence[str] | None, optional): The colors of the channels.
            Defaults to None.
        channel_active (Sequence[bool] | None, optional): Whether the channels are
            active. Defaults to None.
        overwrite (bool, optional): Whether to overwrite an existing image.
            Defaults to True.
        version (NgffVersion, optional): The version of the OME-Zarr specification.
            Defaults to DefaultNgffVersion.
    """
    handler = create_empty_image_container(
        store=store,
        shape=shape,
        pixelsize=xy_pixelsize,
        z_spacing=z_spacing,
        time_spacing=time_spacing,
        levels=levels,
        yx_scaling_factor=xy_scaling_factor,
        z_scaling_factor=z_scaling_factor,
        space_unit=space_unit,
        time_unit=time_unit,
        axes_names=axes_names,
        name=name,
        chunks=chunks,
        dtype=dtype,
        dimension_separator=dimension_separator,
        compressors=compressors,
        overwrite=overwrite,
        version=version,
    )

    ome_zarr = OmeZarrContainer(group_handler=handler)
    ome_zarr.set_channel_meta(
        labels=channel_labels,
        wavelength_id=channel_wavelengths,
        percentiles=None,
        colors=channel_colors,
        active=channel_active,
    )
    return ome_zarr


def create_ome_zarr_from_array(
    store: StoreOrGroup,
    array: np.ndarray,
    xy_pixelsize: float,
    z_spacing: float = 1.0,
    time_spacing: float = 1.0,
    levels: int | list[str] = 5,
    xy_scaling_factor: float = 2.0,
    z_scaling_factor: float = 1.0,
    space_unit: SpaceUnits = DefaultSpaceUnit,
    time_unit: TimeUnits = DefaultTimeUnit,
    axes_names: Sequence[str] | None = None,
    channel_labels: list[str] | None = None,
    channel_wavelengths: list[str] | None = None,
    percentiles: tuple[float, float] | None = (0.1, 99.9),
    channel_colors: Sequence[str] | None = None,
    channel_active: Sequence[bool] | None = None,
    name: str | None = None,
    chunks: Sequence[int] | Literal["auto"] = "auto",
    dimension_separator: Literal[".", "/"] = "/",
    compressors: CompressorLike = "auto",
    overwrite: bool = False,
    version: NgffVersions = DefaultNgffVersion,
) -> OmeZarrContainer:
    """Create an OME-Zarr image from a numpy array.

    Args:
        store (StoreOrGroup): The Zarr store or group to create the image in.
        array (np.ndarray): The image data.
        xy_pixelsize (float): The pixel size in x and y dimensions.
        z_spacing (float, optional): The spacing between z slices. Defaults to 1.0.
        time_spacing (float, optional): The spacing between time points.
            Defaults to 1.0.
        levels (int | list[str], optional): The number of levels in the pyramid or a
            list of level names. Defaults to 5.
        xy_scaling_factor (float, optional): The down-scaling factor in x and y
            dimensions. Defaults to 2.0.
        z_scaling_factor (float, optional): The down-scaling factor in z dimension.
            Defaults to 1.0.
        space_unit (SpaceUnits, optional): The unit of space. Defaults to
            DefaultSpaceUnit.
        time_unit (TimeUnits, optional): The unit of time. Defaults to
            DefaultTimeUnit.
        axes_names (Sequence[str] | None, optional): The names of the axes.
            If None the canonical names are used. Defaults to None.
        name (str | None, optional): The name of the image. Defaults to None.
        chunks (Sequence[int] | None, optional): The chunk shape. If None the shape
            is used. Defaults to None.
        channel_labels (list[str] | None, optional): The labels of the channels.
            Defaults to None.
        channel_wavelengths (list[str] | None, optional): The wavelengths of the
            channels. Defaults to None.
        percentiles (tuple[float, float] | None, optional): The percentiles of the
            channels. Defaults to None.
        channel_colors (Sequence[str] | None, optional): The colors of the channels.
            Defaults to None.
        channel_active (Sequence[bool] | None, optional): Whether the channels are
            active. Defaults to None.
        dimension_separator (DIMENSION_SEPARATOR): The separator to use for
            dimensions. Defaults to "/".
        compressors (CompressorLike): The compressors to use. Defaults to "auto".
        overwrite (bool, optional): Whether to overwrite an existing image.
            Defaults to True.
        version (str, optional): The version of the OME-Zarr specification.
            Defaults to DefaultNgffVersion.
    """
    handler = create_empty_image_container(
        store=store,
        shape=array.shape,
        pixelsize=xy_pixelsize,
        z_spacing=z_spacing,
        time_spacing=time_spacing,
        levels=levels,
        yx_scaling_factor=xy_scaling_factor,
        z_scaling_factor=z_scaling_factor,
        space_unit=space_unit,
        time_unit=time_unit,
        axes_names=axes_names,
        name=name,
        chunks=chunks,
        dtype=str(array.dtype),
        overwrite=overwrite,
        dimension_separator=dimension_separator,
        compressors=compressors,
        version=version,
    )

    ome_zarr = OmeZarrContainer(group_handler=handler)
    image = ome_zarr.get_image()
    image.set_array(array)
    image.consolidate()
    ome_zarr.set_channel_meta(
        labels=channel_labels,
        wavelength_id=channel_wavelengths,
        percentiles=percentiles,
        colors=channel_colors,
        active=channel_active,
    )
    return ome_zarr



import dask.array as da
import numpy as np
import pytest

from ngio.io_pipes._ops_axes import (
    _build_axes_ops,
    get_as_dask_axes_ops,
    get_as_numpy_axes_ops,
    get_as_sequence_axes_ops,
    set_as_dask_axes_ops,
    set_as_numpy_axes_ops,
    set_as_sequence_axes_ops,
)


@pytest.mark.parametrize(
    "input_axes,output_axes,input_shape,output_shape",
    [
        ("tczyx", "zyx", (1, 1, 10, 11, 12), (10, 11, 12)),
        ("czyx", "zcyx", (1, 10, 11, 12), (10, 1, 11, 12)),
        ("zyx", "tczyx", (10, 11, 12), (1, 1, 10, 11, 12)),
        ("czyx", "yxtc", (3, 1, 11, 12), (11, 12, 1, 3)),
    ],
)
def test_axes_base(input_axes, output_axes, input_shape, output_shape):
    input_axes = tuple(input_axes)
    output_axes = tuple(output_axes)
    ops = _build_axes_ops(input_axes, output_axes)

    # Round for tuples
    get_shape = get_as_sequence_axes_ops(input_shape, axes_ops=ops, default=1)
    assert tuple(get_shape) == output_shape
    set_shape = set_as_sequence_axes_ops(get_shape, axes_ops=ops, default=1)
    assert tuple(set_shape) == input_shape

    # Numpy
    input_data = np.zeros(input_shape)
    get_data = get_as_numpy_axes_ops(input_data, axes_ops=ops)
    assert get_data.shape == output_shape
    set_data = set_as_numpy_axes_ops(get_data, axes_ops=ops)
    assert set_data.shape == input_shape

    # Dask
    input_data = da.zeros(input_shape)
    get_data = get_as_dask_axes_ops(input_data, axes_ops=ops)
    assert get_data.shape == output_shape
    set_data = set_as_dask_axes_ops(get_data, axes_ops=ops)
    assert set_data.shape == input_shape
